import typing

from langchain_core.messages import AnyMessage
from typing_extensions import Required, TypedDict


class Prompt(TypedDict, total=False):
    """TypedDict for structured prompt management and optimization.

    Example:
        ```python
        from langmem import Prompt

        prompt = Prompt(
            name="extract_entities",
            prompt="Extract key entities from the text:",
            update_instructions="Make minimal changes, only address where"
            " errors have occurred after reasoning over why they occur.",
            when_to_update="If there seem to be errors in recall of named entities.",
        )
        ```

    The name and prompt fields are required. Optional fields control optimization:
    - update_instructions: Guidelines for modifying the prompt
    - when_to_update: Dependencies between prompts during optimization

    Use in the prompt optimizers.
    """

    name: Required[str]
    prompt: Required[str]
    update_instructions: str | None
    when_to_update: str | None


class AnnotatedTrajectory(typing.NamedTuple):
    """Conversation history (list of messages) with optional feedback for prompt optimization.

    Example:
        ```python
        from langmem.prompts.types import AnnotatedTrajectory

        trajectory = AnnotatedTrajectory(
            messages=[
                {"role": "user", "content": "What pizza is good around here?"},
                {"role": "assistant", "content": "Try LangPizza™️"},
                {"role": "user", "content": "Stop advertising to me."},
                {"role": "assistant", "content": "BUT YOU'LL LOVE IT!"},
            ],
            feedback={
                "developer_feedback": "too pushy",
                "score": 0,
            },
        )
        ```
    """

    messages: typing.Sequence[AnyMessage]
    feedback: dict[str, str | int | bool] | str | None = None


class OptimizerInput(TypedDict):
    """Input for single-prompt optimization.

    Example:
        ```python
        {
            "trajectories": [
                AnnotatedTrajectory(
                    messages=[
                        {"role": "user", "content": "What's the weather like?"},
                        {
                            "role": "assistant",
                            "content": "I'm sorry, I can't tell you that",
                        },
                    ],
                    feedback="Should have checked your search tool.",
                ),
            ],
            "prompt": Prompt(
                name="main_assistant",
                prompt="You are a helpful assistant with a search tool.",
                update_instructions="Make minimal changes, only address where "
                "errors have occurred after reasoning over why they occur.",
                when_to_update="Any time you notice the agent behaving in a way that doesn't help the user.",
            ),
        }
        ```
    """

    trajectories: typing.Sequence[AnnotatedTrajectory] | str
    prompt: str | Prompt


class MultiPromptOptimizerInput(TypedDict):
    """Input for optimizing multiple prompts together, maintaining consistency.

    Example:
        ```python
        {
            "trajectories": [
                AnnotatedTrajectory(
                    messages=[
                        {"role": "user", "content": "Tell me about this image"},
                        {
                            "role": "assistant",
                            "content": "I see a dog playing in a park",
                        },
                        {"role": "user", "content": "What breed is it?"},
                        {
                            "role": "assistant",
                            "content": "Sorry, I can't tell the breed",
                        },
                    ],
                    feedback="Vision model wasn't used for breed detection",
                ),
            ],
            "prompts": [
                Prompt(
                    name="vision_extract",
                    prompt="Extract visual details from the image",
                    update_instructions="Focus on using vision model capabilities",
                ),
                Prompt(
                    name="vision_classify",
                    prompt="Classify specific attributes in the image",
                    when_to_update="After vision_extract is updated",
                ),
            ],
        }
        ```
    """

    trajectories: typing.Sequence[AnnotatedTrajectory] | str
    prompts: list[Prompt]



import asyncio
import typing

import langsmith as ls
from langchain_core.language_models import BaseChatModel
from langchain_core.runnables import Runnable, RunnableConfig
from pydantic import BaseModel, Field, model_validator
from trustcall import create_extractor

import langmem.utils as utils
from langmem.prompts import types as prompt_types
from langmem.prompts.gradient import (
    GradientOptimizerConfig,
    create_gradient_prompt_optimizer,
)
from langmem.prompts.metaprompt import (
    MetapromptOptimizerConfig,
    create_metaprompt_optimizer,
)
from langmem.prompts.stateless import PromptMemoryMultiple
from langmem.prompts.types import Prompt

KINDS = typing.Literal["gradient", "metaprompt", "prompt_memory"]


@typing.overload
def create_prompt_optimizer(
    model: str | BaseChatModel,
    kind: typing.Literal["gradient"] = "gradient",
    config: typing.Optional[GradientOptimizerConfig] = None,
) -> Runnable[prompt_types.OptimizerInput, str]: ...


@typing.overload
def create_prompt_optimizer(
    model: str | BaseChatModel,
    kind: typing.Literal["metaprompt"] = "metaprompt",
    config: typing.Optional[MetapromptOptimizerConfig] = None,
) -> Runnable[prompt_types.OptimizerInput, str]: ...


@typing.overload
def create_prompt_optimizer(
    model: str | BaseChatModel,
    kind: typing.Literal["prompt_memory"] = "prompt_memory",
    config: None = None,
) -> Runnable[prompt_types.OptimizerInput, str]: ...


def create_prompt_optimizer(
    model: str | BaseChatModel,
    /,
    *,
    kind: KINDS = "gradient",
    config: typing.Union[
        GradientOptimizerConfig, MetapromptOptimizerConfig, None
    ] = None,
) -> Runnable[prompt_types.OptimizerInput, str]:
    """Create a prompt optimizer that improves prompt effectiveness.

    This function creates an optimizer that can analyze and improve prompts for better
    performance with language models. It supports multiple optimization strategies to
    iteratively enhance prompt quality and effectiveness.

    Args:
        model (Union[str, BaseChatModel]): The language model to use for optimization.
            Can be a model name string or a BaseChatModel instance.
        kind (Literal["gradient", "prompt_memory", "metaprompt"]): The optimization
            strategy to use. Each strategy offers different benefits:

            - gradient: Separates concerns between finding areas for improvement
                and recommending updates
            - prompt_memory: Simple single-shot metaprompt
            - metaprompt: Supports reflection but each step is a single LLM call.
        config (Optional[OptimizerConfig]): Configuration options for the optimizer.
            The type depends on the chosen strategy:

                - GradientOptimizerConfig for kind="gradient"
                - PromptMemoryConfig for kind="prompt_memory"
                - MetapromptOptimizerConfig for kind="metaprompt"
            Defaults to None.

    Returns:
        optimizer (Runnable[prompt_types.OptimizerInput, str]): A callable that takes conversation trajectories and/or prompts and returns optimized versions.

    ## Optimization Strategies

    ### 1. Gradient Optimizer
    ```mermaid
    sequenceDiagram
        participant U as User
        participant O as Optimizer
        participant R as Reflection
        participant U2 as Update

        U->>O: Prompt + Feedback
        loop For min_steps to max_steps
            O->>R: Think/Critique Current State
            R-->>O: Proposed Improvements
            O->>U2: Apply Update
            U2-->>O: Updated Prompt
        end
        O->>U: Final Optimized Prompt
    ```

    The gradient optimizer uses reflection to propose improvements:

    1. Analyzes prompt and feedback through reflection cycles
    2. Proposes specific improvements
    3. Applies single-step updates

    Configuration (GradientOptimizerConfig):

    - gradient_prompt: Custom prompt for predicting "what to improve"
    - metaprompt: Custom prompt for applying the improvements
    - max_reflection_steps: Maximum reflection iterations (default: 3)
    - min_reflection_steps: Minimum reflection iterations (default: 1)

    ### 2. Meta-Prompt Optimizer
    ```mermaid
    sequenceDiagram
        participant U as User
        participant M as MetaOptimizer
        participant A as Analysis
        participant U2 as Update

        U->>M: Prompt + Examples
        M->>A: Analyze Examples
        A-->>M: Proposed Update
        M->>U2: Apply Update
        U2-->>U: Enhanced Prompt
    ```

    Uses meta-learning to directly propose updates:

    1. Analyzes examples to understand patterns
    2. Proposes direct prompt updates
    3. Applies updates in a single step

    Configuration (MetapromptOptimizerConfig):

    - metaprompt: Custom instructions on how to update the prompt
    - max_reflection_steps: Maximum meta-learning steps (default: 3)
    - min_reflection_steps: Minimum meta-learning steps (default: 1)

    ### 3. Prompt Memory Optimizer
    ```mermaid
    sequenceDiagram
        participant U as User
        participant P as PromptMemory
        participant M as Memory

        U->>P: Prompt + History
        P->>M: Extract Patterns
        M-->>P: Success Patterns
        P->>U: Updated Prompt
    ```

    Learns from conversation history:

    1. Extracts successful patterns from past interactions
    2. Identifies improvement areas from feedback
    3. Applies learned patterns to new prompts

    No additional configuration required.

    !!! example "Examples"
        Basic prompt optimization:
        ```python
        from langmem import create_prompt_optimizer

        optimizer = create_prompt_optimizer("anthropic:claude-3-5-sonnet-latest")

        # Example conversation with feedback
        conversation = [
            {"role": "user", "content": "Tell me about the solar system"},
            {"role": "assistant", "content": "The solar system consists of..."},
        ]
        feedback = {"clarity": "needs more structure"}

        # Use conversation history to improve the prompt
        trajectories = [(conversation, feedback)]
        better_prompt = await optimizer.ainvoke(
            {"trajectories": trajectories, "prompt": "You are an astronomy expert"}
        )
        print(better_prompt)
        # Output: 'Provide a comprehensive overview of the solar system...'
        ```

        Optimizing with conversation feedback:
        ```python
        from langmem import create_prompt_optimizer

        optimizer = create_prompt_optimizer(
            "anthropic:claude-3-5-sonnet-latest", kind="prompt_memory"
        )

        # Conversation with feedback about what could be improved
        conversation = [
            {"role": "user", "content": "How do I write a bash script?"},
            {"role": "assistant", "content": "Let me explain bash scripting..."},
        ]
        feedback = "Response should include a code example"

        # Use the conversation and feedback to improve the prompt
        trajectories = [(conversation, {"feedback": feedback})]
        better_prompt = await optimizer(trajectories, "You are a coding assistant")
        print(better_prompt)
        # Output: 'You are a coding assistant that always includes...'
        ```

        Meta-prompt optimization for complex tasks:
        ```python
        from langmem import create_prompt_optimizer

        optimizer = create_prompt_optimizer(
            "anthropic:claude-3-5-sonnet-latest",
            kind="metaprompt",
            config={"max_reflection_steps": 3, "min_reflection_steps": 1},
        )

        # Complex conversation that needs better structure
        conversation = [
            {"role": "user", "content": "Explain quantum computing"},
            {"role": "assistant", "content": "Quantum computing uses..."},
        ]
        feedback = "Need better organization and concrete examples"

        # Optimize with meta-learning
        trajectories = [(conversation, feedback)]
        improved_prompt = await optimizer(
            trajectories, "You are a quantum computing expert"
        )
        ```

    !!! warning "Performance Considerations"

        Each strategy has different LLM call patterns:

        - prompt_memory: 1 LLM call total
            - Fastest as it only needs one pass
        - metaprompt: 1-5 LLM calls (configurable)
            - Each step is one LLM call
            - Default range: min 2, max 5 reflection steps
        - gradient: 2-10 LLM calls (configurable)
            - Each step requires 2 LLM calls (think + critique)
            - Default range: min 2, max 5 reflection steps

    !!! tip "Strategy Selection"
        Choose based on your needs:

        1. Prompt Memory: Simplest prompting strategy
            - Limited ability to learn from complicated patterns
        2. Metaprompt: Balance of speed and improvement
            - Moderate cost (2-5 LLM calls)
        3. Gradient: Most thorough but expensive
            - Highest cost (4-10 LLM calls)
            - Uses separation of concerns to extract feedback from more conversational context.
    """
    if kind == "gradient":
        return create_gradient_prompt_optimizer(model, config)  # type: ignore
    elif kind == "metaprompt":
        return create_metaprompt_optimizer(model, config)  # type: ignore
    elif kind == "prompt_memory":
        return PromptMemoryMultiple(model)  # type: ignore
    else:
        raise NotImplementedError(
            f"Unsupported optimizer kind: {kind}.\nExpected one of {KINDS}"
        )


class MultiPromptOptimizer(
    Runnable[prompt_types.MultiPromptOptimizerInput, list[Prompt]]
):
    def __init__(
        self,
        model: str | BaseChatModel,
        /,
        *,
        kind: typing.Literal["gradient", "prompt_memory", "metaprompt"] = "gradient",
        config: typing.Optional[dict] = None,
    ):
        self.model = model
        self.kind = kind
        self.config = config
        # Build a single-prompt optimizer used internally
        self._optimizer = create_prompt_optimizer(model, kind=kind, config=config)

    async def ainvoke(
        self,
        input: prompt_types.MultiPromptOptimizerInput,
        config: typing.Optional[RunnableConfig] = None,
        **kwargs: typing.Any,
    ) -> list[Prompt]:
        async with ls.trace(
            name="multi_prompt_optimizer.ainvoke",
            inputs=input,
            metadata={"kind": self.kind},
        ) as rt:
            trajectories = input["trajectories"]
            prompts = input["prompts"]

            # Get available prompt names.
            choices = [p["name"] for p in prompts]
            sessions_str = (
                trajectories
                if isinstance(trajectories, str)
                else utils.format_sessions(trajectories)
            )

            # If only one prompt and no explicit when_to_update instruction, simply update it.
            if len(prompts) == 1 and prompts[0].get("when_to_update") is None:
                updated_prompt = await self._optimizer(trajectories, prompts[0])
                rt.add_outputs({"output": [{**prompts[0], "prompt": updated_prompt}]})
                return [{**prompts[0], "prompt": updated_prompt}]

            class Classify(BaseModel):
                """After analyzing the provided trajectories, determine which prompt modules (if any) contributed to degraded performance."""

                reasoning: str = Field(
                    description="Reasoning for which prompts to update."
                )
                which: list[str] = Field(
                    description=f"List of prompt names that should be updated. Must be among {choices}"
                )

                @model_validator(mode="after")
                def validate_choices(self) -> "Classify":
                    invalid = set(self.which) - set(choices)
                    if invalid:
                        raise ValueError(
                            f"Invalid choices: {invalid}. Must be among: {choices}"
                        )
                    return self

            classifier = create_extractor(
                self.model, tools=[Classify], tool_choice="Classify"
            )
            prompt_joined_content = "".join(
                f"{p['name']}: {p['prompt']}\n" for p in prompts
            )
            classification_prompt = f"""Analyze the following trajectories and decide which prompts 
ought to be updated to improve the performance on future trajectories:

{sessions_str}

Below are the prompts being optimized:
{prompt_joined_content}

Return JSON with "which": [...], listing the names of prompts that need updates."""
            result = await classifier.ainvoke(classification_prompt)
            to_update = result["responses"][0].which  # type: ignore

            which_to_update = [p for p in prompts if p["name"] in to_update]

            # Update each chosen prompt concurrently.
            updated_results = await asyncio.gather(
                *[self._optimizer(trajectories, prompt=p) for p in which_to_update]
            )
            updated_map = {
                p["name"]: new_text
                for p, new_text in zip(which_to_update, updated_results)
            }

            # Merge updates back into the prompt list.
            final_list = []
            for p in prompts:
                if p["name"] in updated_map:
                    final_list.append({**p, "prompt": updated_map[p["name"]]})
                else:
                    final_list.append(p)
            rt.add_outputs({"output": final_list})
            return final_list

    def invoke(
        self,
        input: prompt_types.MultiPromptOptimizerInput,
        config: typing.Optional[RunnableConfig] = None,
        **kwargs: typing.Any,
    ) -> list[Prompt]:
        with ls.trace(
            name="multi_prompt_optimizer.invoke",
            inputs=input,
            metadata={"kind": self.kind},
        ) as rt:
            trajectories = input["trajectories"]
            prompts = input["prompts"]

            choices = [p["name"] for p in prompts]
            sessions_str = (
                trajectories
                if isinstance(trajectories, str)
                else utils.format_sessions(trajectories)
            )

            if len(prompts) == 1 and prompts[0].get("when_to_update") is None:
                updated_prompt = self._optimizer.invoke(
                    {"trajectories": trajectories, "prompt": prompts[0]}
                )
                result = [{**prompts[0], "prompt": updated_prompt}]
                rt.add_outputs({"output": result})
                return typing.cast(list[Prompt], result)

            class Classify(BaseModel):
                """After analyzing the provided trajectories, determine which prompt modules (if any) contributed to degraded performance."""

                reasoning: str = Field(
                    description="Reasoning for which prompts to update."
                )
                which: list[str] = Field(
                    description=f"List of prompt names that should be updated. Must be among {choices}"
                )

                @model_validator(mode="after")
                def validate_choices(self) -> "Classify":
                    invalid = set(self.which) - set(choices)
                    if invalid:
                        raise ValueError(
                            f"Invalid choices: {invalid}. Must be among: {choices}"
                        )
                    return self

            classifier = create_extractor(
                self.model, tools=[Classify], tool_choice="Classify"
            )
            prompt_joined_content = "".join(
                f"{p['name']}: {p['prompt']}\n" for p in prompts
            )
            classification_prompt = f"""Analyze the following trajectories and decide which prompts 
ought to be updated to improve the performance on future trajectories:

{sessions_str}

Below are the prompts being optimized:
{prompt_joined_content}

Return JSON with "which": [...], listing the names of prompts that need updates."""
            result = classifier.invoke(classification_prompt)
            to_update = result["responses"][0].which  # type: ignore

            which_to_update = [p for p in prompts if p["name"] in to_update]
            updated_map = {}
            for p in which_to_update:
                updated_text = self._optimizer.invoke(
                    {"trajectories": trajectories, "prompt": p}
                )
                updated_map[p["name"]] = updated_text

            final_list = []
            for p in prompts:
                if p["name"] in updated_map:
                    final_list.append({**p, "prompt": updated_map[p["name"]]})
                else:
                    final_list.append(p)
            rt.add_outputs({"output": final_list})
            return final_list

    async def __call__(
        self,
        trajectories: typing.Sequence[prompt_types.AnnotatedTrajectory] | str,
        prompts: list[Prompt],
    ) -> list[Prompt]:
        """Allow calling the object like: await optimizer(trajectories, prompts)"""
        return await self.ainvoke(
            prompt_types.MultiPromptOptimizerInput(
                trajectories=trajectories, prompts=prompts
            )
        )


def create_multi_prompt_optimizer(
    model: str | BaseChatModel,
    /,
    *,
    kind: typing.Literal["gradient", "prompt_memory", "metaprompt"] = "gradient",
    config: typing.Optional[dict] = None,
) -> Runnable[prompt_types.MultiPromptOptimizerInput, list[Prompt]]:
    """Create a multi-prompt optimizer that improves prompt effectiveness.

    This function creates an optimizer that can analyze and improve multiple prompts
    simultaneously using the same optimization strategy. Each prompt is optimized using
    the selected strategy (see `create_prompt_optimizer` for strategy details).

    Args:
        model (Union[str, BaseChatModel]): The language model to use for optimization.
            Can be a model name string or a BaseChatModel instance.
        kind (Literal["gradient", "prompt_memory", "metaprompt"]): The optimization
            strategy to use. Each strategy offers different benefits:
            - gradient: Iteratively improves through reflection
            - prompt_memory: Uses successful past prompts
            - metaprompt: Learns optimal patterns via meta-learning
            Defaults to "gradient".
        config (Optional[OptimizerConfig]): Configuration options for the optimizer.
            The type depends on the chosen strategy:
                - GradientOptimizerConfig for kind="gradient"
                - PromptMemoryConfig for kind="prompt_memory"
                - MetapromptOptimizerConfig for kind="metaprompt"
            Defaults to None.

    Returns:
        MultiPromptOptimizer: A Runnable that takes conversation trajectories and prompts
            and returns optimized versions.

    ```mermaid
    sequenceDiagram
        participant U as User
        participant M as Multi-prompt Optimizer
        participant C as Credit Assigner
        participant O as Single-prompt Optimizer
        participant P as Prompts

        U->>M: Annotated Trajectories + Prompts
        activate M
        Note over M: Using pre-initialized<br/>single-prompt optimizer

        M->>C: Analyze trajectories
        activate C
        Note over C: Determine which prompts<br/>need improvement
        C-->>M: Credit assignment results
        deactivate C

        loop For each prompt needing update
            M->>O: Optimize prompt
            activate O
            O->>P: Apply optimization strategy
            Note over O,P: Gradient/Memory/Meta<br/>optimization
            P-->>O: Optimized prompt
            O-->>M: Return result
            deactivate O
        end

        M->>U: Return optimized prompts
        deactivate M
    ```

    The system optimizer:

    !!! example "Examples"
        Basic prompt optimization:
        ```python
        from langmem import create_multi_prompt_optimizer

        optimizer = create_multi_prompt_optimizer("anthropic:claude-3-5-sonnet-latest")

        # Example conversation with feedback
        conversation = [
            {"role": "user", "content": "Tell me about the solar system"},
            {"role": "assistant", "content": "The solar system consists of..."},
        ]
        feedback = {"clarity": "needs more structure"}

        # Use conversation history to improve the prompts
        trajectories = [(conversation, feedback)]
        prompts = [
            {"name": "research", "prompt": "Research the given topic thoroughly"},
            {"name": "summarize", "prompt": "Summarize the research findings"},
        ]
        better_prompts = await optimizer.ainvoke(
            {"trajectories": trajectories, "prompts": prompts}
        )
        print(better_prompts)
        ```

        Optimizing with conversation feedback:
        ```python
        from langmem import create_multi_prompt_optimizer

        optimizer = create_multi_prompt_optimizer(
            "anthropic:claude-3-5-sonnet-latest", kind="prompt_memory"
        )

        # Conversation with feedback about what could be improved
        conversation = [
            {"role": "user", "content": "How do I write a bash script?"},
            {"role": "assistant", "content": "Let me explain bash scripting..."},
        ]
        feedback = "Response should include a code example"

        # Use the conversation and feedback to improve the prompts
        trajectories = [(conversation, {"feedback": feedback})]
        prompts = [
            {"name": "explain", "prompt": "Explain the concept"},
            {"name": "example", "prompt": "Provide a practical example"},
        ]
        better_prompts = await optimizer(trajectories, prompts)
        ```

        Controlling the max number of reflection steps:
        ```python
        from langmem import create_multi_prompt_optimizer

        optimizer = create_multi_prompt_optimizer(
            "anthropic:claude-3-5-sonnet-latest",
            kind="metaprompt",
            config={"max_reflection_steps": 3, "min_reflection_steps": 1},
        )

        # Complex conversation that needs better structure
        conversation = [
            {"role": "user", "content": "Explain quantum computing"},
            {"role": "assistant", "content": "Quantum computing uses..."},
        ]
        # Explicit feedback is optional
        feedback = None

        # Optimize with meta-learning
        trajectories = [(conversation, feedback)]
        prompts = [
            {"name": "concept", "prompt": "Explain quantum concepts"},
            {"name": "application", "prompt": "Show practical applications"},
            {"name": "example", "prompt": "Give concrete examples"},
        ]
        improved_prompts = await optimizer(trajectories, prompts)
        ```
    """
    return MultiPromptOptimizer(model, kind=kind, config=config)


__all__ = ["create_prompt_optimizer", "create_multi_prompt_optimizer"]



from dataclasses import dataclass
from typing import Any, Callable, Iterable, cast

from langchain_core.language_models import LanguageModelLike
from langchain_core.messages import (
    AIMessage,
    AnyMessage,
    HumanMessage,
    MessageLikeRepresentation,
    SystemMessage,
)
from langchain_core.messages.utils import count_tokens_approximately
from langchain_core.prompts.chat import ChatPromptTemplate, ChatPromptValue
from langgraph.utils.runnable import RunnableCallable
from pydantic import BaseModel

TokenCounter = Callable[[Iterable[MessageLikeRepresentation]], int]


DEFAULT_INITIAL_SUMMARY_PROMPT = ChatPromptTemplate.from_messages(
    [
        ("placeholder", "{messages}"),
        ("user", "Create a summary of the conversation above:"),
    ]
)


DEFAULT_EXISTING_SUMMARY_PROMPT = ChatPromptTemplate.from_messages(
    [
        ("placeholder", "{messages}"),
        (
            "user",
            "This is summary of the conversation so far: {existing_summary}\n\n"
            "Extend this summary by taking into account the new messages above:",
        ),
    ]
)

DEFAULT_FINAL_SUMMARY_PROMPT = ChatPromptTemplate.from_messages(
    [
        # if exists
        ("placeholder", "{system_message}"),
        ("system", "Summary of the conversation so far: {summary}"),
        ("placeholder", "{messages}"),
    ]
)


@dataclass
class RunningSummary:
    """Object for storing information about the previous summarization.

    Used on subsequent calls to summarize_messages to avoid summarizing the same messages.
    """

    summary: str
    """Latest summary of the messages, updated every time the summarization is performed."""

    summarized_message_ids: set[str]
    """The IDs of all of the messages that have been previously summarized."""


@dataclass
class SummarizationResult:
    """Result of message summarization."""

    messages: list[AnyMessage]
    """List of updated messages that are ready to be input to the LLM after summarization, including a message with a summary (if any)."""

    running_summary: RunningSummary | None = None
    """Information about previous summarization (the summary and the IDs of the previously summarized messages.
    Can be None if no summarization was performed (not enough messages to summarize).
    """


def summarize_messages(
    messages: list[AnyMessage],
    *,
    running_summary: RunningSummary | None,
    model: LanguageModelLike,
    max_tokens: int,
    max_summary_tokens: int = 256,
    token_counter: TokenCounter = count_tokens_approximately,
    initial_summary_prompt: ChatPromptTemplate = DEFAULT_INITIAL_SUMMARY_PROMPT,
    existing_summary_prompt: ChatPromptTemplate = DEFAULT_EXISTING_SUMMARY_PROMPT,
    final_prompt: ChatPromptTemplate = DEFAULT_FINAL_SUMMARY_PROMPT,
) -> SummarizationResult:
    """Summarize messages when they exceed a token limit and replace them with a summary message.

    This function processes the messages from oldest to newest: once the cumulative number of message tokens
    reaches `max_tokens`, all messages within `max_tokens` are summarized (excluding the system message, if any)
    and replaced with a new summary message. The resulting list of messages is [summary_message] + remaining_messages.

    Args:
        messages: The list of messages to process.
        running_summary: Optional running summary object with information about the previous summarization. If provided:
            - only messages that were **not** previously summarized will be processed
            - if no new summary is generated, the running summary will be added to the returned messages
            - if a new summary needs to be generated, it is generated by incorporating the existing summary value from the running summary
        model: The language model to use for generating summaries.
        max_tokens: Maximum number of tokens to return.
            Will also be used as a threshold for triggering the summarization: once the cumulative number of message tokens
            reaches `max_tokens`, all messages within `max_tokens` will be summarized.

            !!! Note

                If the last message within `max_tokens` is an AI message with tool calls or a human message,
                this message will not be summarized, and instead will be added to the returned messages.
        max_summary_tokens: Maximum number of tokens to budget for the summary.

            !!! Note

                This parameter is not passed to the summary-generating LLM to limit the length of the summary.
                It is only used for correctly estimating the maximum allowed token budget.
                If you want to enforce it, you would need to pass `model.bind(max_tokens=max_summary_tokens)`
                as the `model` parameter to this function.
        token_counter: Function to count tokens in a message. Defaults to approximate counting.
            For more accurate counts you can use `model.get_num_tokens_from_messages`.
        initial_summary_prompt: Prompt template for generating the first summary.
        existing_summary_prompt: Prompt template for updating an existing (running) summary.
        final_prompt: Prompt template that combines summary with the remaining messages before returning.

    Returns:
        A SummarizationResult object containing the updated messages and a running summary.
            - messages: list of updated messages ready to be input to the LLM
            - running_summary: RunningSummary object
                - summary: text of the latest summary
                - summarized_message_ids: set of message IDs that were previously summarized

    Example:
        ```pycon
        >>> from langgraph.graph import StateGraph, START, MessagesState
        >>> from langgraph.checkpoint.memory import InMemorySaver
        >>> from langmem.short_term import summarize_messages, RunningSummary
        >>> from langchain_openai import ChatOpenAI

        >>> model = ChatOpenAI(model="gpt-4o")
        >>> summarization_model = model.bind(max_tokens=128)

        >>> class SummaryState(MessagesState):
        ...     summary: RunningSummary | None

        >>> def call_model(state):
        ...     summarization_result = summarize_messages(
        ...         state["messages"],
        ...         running_summary=state.get("summary"),
        ...         model=summarization_model,
        ...         max_tokens=256,
        ...         max_summary_tokens=128
        ...     )
        ...     response = model.invoke(summarization_result.messages)
        ...     state_update = {"messages": [response]}
        ...     if summarization_result.running_summary:
        ...         state_update["summary"] = summarization_result.running_summary
        ...     return state_update

        >>> checkpointer = InMemorySaver()
        >>> workflow = StateGraph(SummaryState)
        >>> workflow.add_node(call_model)
        >>> workflow.add_edge(START, "call_model")
        >>> graph = workflow.compile(checkpointer=checkpointer)

        >>> config = {"configurable": {"thread_id": "1"}}
        >>> graph.invoke({"messages": "hi, my name is bob"}, config)
        >>> graph.invoke({"messages": "write a short poem about cats"}, config)
        >>> graph.invoke({"messages": "now do the same but for dogs"}, config)
        >>> graph.invoke({"messages": "what's my name?"}, config)
        ```
    """
    if max_summary_tokens >= max_tokens:
        raise ValueError("`max_summary_tokens` must be less than `max_tokens`.")

    max_tokens_to_summarize = max_tokens
    # Adjust the remaining token budget to account for the summary to be added
    max_remaining_tokens = max_tokens - max_summary_tokens
    # First handle system message if present
    if messages and isinstance(messages[0], SystemMessage):
        existing_system_message = messages[0]
        # remove the system message from the list of messages to summarize
        messages = messages[1:]
        # adjust the remaining token budget to account for the system message to be re-added
        max_remaining_tokens -= token_counter([existing_system_message])
    else:
        existing_system_message = None

    if not messages:
        return SummarizationResult(
            running_summary=running_summary,
            messages=(
                messages
                if existing_system_message is None
                else [existing_system_message] + messages
            ),
        )

    # Get previously summarized messages, if any
    summarized_message_ids = set()
    if running_summary:
        summarized_message_ids = running_summary.summarized_message_ids
        # Adjust the summarization token budget to account for the previous summary
        max_tokens_to_summarize -= token_counter(
            [SystemMessage(content=running_summary.summary)]
        )

    total_summarized_messages = len(summarized_message_ids)

    # Go through messages to count tokens and find cutoff point
    n_tokens = 0
    idx = max(0, total_summarized_messages - 1)
    # We need to output messages that fit within max_tokens.
    # Assuming that the summarization LLM also needs at most max_tokens
    # that will be turned into at most max_summary_tokens, you can try
    # to process at most max_tokens * 2 - max_summary_tokens
    max_total_tokens = max_tokens + max_remaining_tokens
    for i in range(total_summarized_messages, len(messages)):
        message = messages[i]
        if message.id is None:
            raise ValueError("Messages are required to have ID field.")

        if message.id in summarized_message_ids:
            raise ValueError(
                f"Message with ID {message.id} has already been summarized."
            )

        n_tokens += token_counter([message])

        # If we're still under max_tokens_to_summarize, update the potential cutoff point
        if n_tokens <= max_tokens_to_summarize:
            idx = i

        # Check if we've exceeded the absolute maximum
        if n_tokens > max_total_tokens:
            raise ValueError(
                f"`summarize_messages` cannot handle more than {max_total_tokens} tokens: "
                f"resulting message history will exceed max_tokens limit ({max_tokens}). "
                "Please adjust `max_tokens` / `max_summary_tokens` or decrease the input size."
            )

    # If we haven't exceeded max_tokens, we don't need to summarize
    # Note: we don't return here since we might still need to include the existing summary
    if n_tokens <= max_tokens:
        messages_to_summarize = []
    else:
        messages_to_summarize = messages[total_summarized_messages : idx + 1]

    # If the last message is:
    # (1) an AI message with tool calls - remove it
    #   to avoid issues w/ the LLM provider (as it will lack a corresponding tool message)
    # (2) a human message - remove it,
    #   since it is a user input and it doesn't make sense to summarize it without a corresponding AI message
    while messages_to_summarize and (
        (
            isinstance(messages_to_summarize[-1], AIMessage)
            and messages_to_summarize[-1].tool_calls
        )
        or isinstance(messages_to_summarize[-1], HumanMessage)
    ):
        messages_to_summarize.pop()

    if messages_to_summarize:
        if running_summary:
            summary_messages = cast(
                ChatPromptValue,
                existing_summary_prompt.invoke(
                    {
                        "messages": messages_to_summarize,
                        "existing_summary": running_summary.summary,
                    }
                ),
            )
        else:
            summary_messages = cast(
                ChatPromptValue,
                initial_summary_prompt.invoke({"messages": messages_to_summarize}),
            )

        summary_response = model.invoke(summary_messages.messages)
        summarized_message_ids = summarized_message_ids | set(
            message.id for message in messages_to_summarize
        )
        total_summarized_messages += len(messages_to_summarize)
        running_summary = RunningSummary(
            summary=summary_response.content,
            summarized_message_ids=summarized_message_ids,
        )

    if running_summary:
        updated_messages = cast(
            ChatPromptValue,
            final_prompt.invoke(
                {
                    "system_message": [existing_system_message]
                    if existing_system_message
                    else [],
                    "summary": running_summary.summary,
                    "messages": messages[total_summarized_messages:],
                }
            ),
        )
        return SummarizationResult(
            running_summary=running_summary,
            messages=updated_messages.messages,
        )
    else:
        # no changes are needed
        return SummarizationResult(
            running_summary=None,
            messages=(
                messages
                if existing_system_message is None
                else [existing_system_message] + messages
            ),
        )


class SummarizationNode(RunnableCallable):
    """A LangGraph node that summarizes messages when they exceed a token limit and replaces them with a summary message."""

    def __init__(
        self,
        *,
        model: LanguageModelLike,
        max_tokens: int,
        max_summary_tokens: int = 256,
        token_counter: TokenCounter = count_tokens_approximately,
        initial_summary_prompt: ChatPromptTemplate = DEFAULT_INITIAL_SUMMARY_PROMPT,
        existing_summary_prompt: ChatPromptTemplate = DEFAULT_EXISTING_SUMMARY_PROMPT,
        final_prompt: ChatPromptTemplate = DEFAULT_FINAL_SUMMARY_PROMPT,
        input_messages_key: str = "messages",
        output_messages_key: str = "summarized_messages",
        name: str = "summarization",
    ) -> None:
        """A LangGraph node that summarizes messages when they exceed a token limit and replaces them with a summary message.

        Processes the messages from oldest to newest: once the cumulative number of message tokens
        reaches `max_tokens`, all messages within `max_tokens` are summarized (excluding the system message, if any)
        and replaced with a new summary message. The resulting list of messages is [summary_message] + remaining_messages.

        Args:
            model: The language model to use for generating summaries.
            max_tokens: Maximum number of tokens to return.
                Will also be used as a threshold for triggering the summarization: once the cumulative number of message tokens
                reaches `max_tokens`, all messages within `max_tokens` will be summarized.

                !!! Note

                    If the last message within max_tokens is an AI message with tool calls or a human message,
                    this message will not be summarized, and instead will be added to the returned messages.
            max_summary_tokens: Maximum number of tokens to budget for the summary.

                !!! Note

                    This parameter is not passed to the summary-generating LLM to limit the length of the summary.
                    It is only used for correctly estimating the maximum allowed token budget.
                    If you want to enforce it, you would need to pass `model.bind(max_tokens=max_summary_tokens)`
                    as the `model` parameter to this function.
            token_counter: Function to count tokens in a message. Defaults to approximate counting.
                For more accurate counts you can use `model.get_num_tokens_from_messages`.
            initial_summary_prompt: Prompt template for generating the first summary.
            existing_summary_prompt: Prompt template for updating an existing (running) summary.
            final_prompt: Prompt template that combines summary with the remaining messages before returning.
            input_messages_key: Key in the input graph state that contains the list of messages to summarize.
            output_messages_key: Key in the state update that contains the list of updated messages.
                !!! Note

                    `output_messages_key` **must be different** from the `input_messages_key`.
                    This is done to decouple summarized messages from the main list of messages in the graph state (i.e., `input_messages_key`).
                    If you want to update / overwrite the main list of messages, you would need to use `summarize_messages` function directly or wrap
                    the invocation of this node in a different node.

            name: Name of the summarization node.

        Returns:
            LangGraph state update in the following format:
                ```
                {
                    output_messages_key: <list of updated messages ready to be input to the LLM after summarization, including a message with a summary (if any)>,
                    "context": {"running_summary": <RunningSummary object>}
                }
                ```

        Example:
            ```pycon
            >>> from typing import Any, TypedDict
            >>> from langchain_openai import ChatOpenAI
            >>> from langchain_core.messages import AnyMessage
            >>> from langgraph.graph import StateGraph, START, MessagesState
            >>> from langgraph.checkpoint.memory import InMemorySaver
            >>> from langmem.short_term import SummarizationNode, RunningSummary
            >>>
            >>> model = ChatOpenAI(model="gpt-4o")
            >>> summarization_model = model.bind(max_tokens=128)
            >>>
            >>> class State(MessagesState):
            ...     context: dict[str, Any]
            ...
            >>> class LLMInputState(TypedDict):
            ...     summarized_messages: list[AnyMessage]
            ...     context: dict[str, Any]
            ...
            >>> summarization_node = SummarizationNode(
            ...     model=summarization_model,
            ...     max_tokens=256,
            ...     max_summary_tokens=128,
            ... )
            >>>
            >>> def call_model(state: LLMInputState):
            ...     response = model.invoke(state["summarized_messages"])
            ...     return {"messages": [response]}
            ...
            >>> checkpointer = InMemorySaver()
            >>> workflow = StateGraph(State)
            >>> workflow.add_node(call_model)
            >>> workflow.add_node("summarize", summarization_node)
            >>> workflow.add_edge(START, "summarize")
            >>> workflow.add_edge("summarize", "call_model")
            >>> graph = workflow.compile(checkpointer=checkpointer)
            >>>
            >>> config = {"configurable": {"thread_id": "1"}}
            >>> graph.invoke({"messages": "hi, my name is bob"}, config)
            >>> graph.invoke({"messages": "write a short poem about cats"}, config)
            >>> graph.invoke({"messages": "now do the same but for dogs"}, config)
            >>> graph.invoke({"messages": "what's my name?"}, config)
            ```
        """
        super().__init__(self._func, name=name, trace=False)
        self.model = model
        self.max_tokens = max_tokens
        self.max_summary_tokens = max_summary_tokens
        self.token_counter = token_counter
        self.initial_summary_prompt = initial_summary_prompt
        self.existing_summary_prompt = existing_summary_prompt
        self.final_prompt = final_prompt
        if input_messages_key == output_messages_key:
            raise ValueError(
                "`input_messages_key` and `output_messages_key` must be different."
            )

        self.input_messages_key = input_messages_key
        self.output_messages_key = output_messages_key

    def _func(self, input: dict[str, Any] | BaseModel) -> dict[str, Any]:
        if isinstance(input, dict):
            messages = input.get(self.input_messages_key)
            context = input.get("context", {})
        elif isinstance(input, BaseModel):
            messages = getattr(input, self.input_messages_key, None)
            context = getattr(input, "context", {})
        else:
            raise ValueError(f"Invalid input type: {type(input)}")

        if messages is None:
            raise ValueError(
                f"Missing required field `{self.input_messages_key}` in the input."
            )

        summarization_result = summarize_messages(
            messages,
            running_summary=context.get("running_summary"),
            model=self.model,
            max_tokens=self.max_tokens,
            max_summary_tokens=self.max_summary_tokens,
            token_counter=self.token_counter,
            initial_summary_prompt=self.initial_summary_prompt,
            existing_summary_prompt=self.existing_summary_prompt,
            final_prompt=self.final_prompt,
        )

        state_update = {self.output_messages_key: summarization_result.messages}
        if summarization_result.running_summary:
            state_update["context"] = {
                **context,
                "running_summary": summarization_result.running_summary,
            }
        return state_update



import re
import typing

from langchain_core.messages import BaseMessage
from pydantic import BaseModel, Field, model_validator

from langmem import utils


def _get_msg_title_repr(title: str) -> str:
    """Get a title representation for a message.

    Args:
        title: The title.
        bold: Whether to bold the title. Default is False.

    Returns:
        The title representation.
    """
    padded = " " + title + " "
    sep_len = (80 - len(padded)) // 2
    sep = "=" * sep_len
    second_sep = sep + "=" if len(padded) % 2 else sep
    return f"{sep}{padded}{second_sep}"


def get_trajectory_clean(messages):
    response = []
    for m in messages:
        if isinstance(m, BaseMessage):
            response.append(m.pretty_repr())
        elif isinstance(m, dict) and "role" in m and "content" in m:
            title = _get_msg_title_repr(m["role"])
            name = m.get("name")
            # TODO: handle non-string content.
            if name is not None:
                title += f"\nName: {name}"
            response.append(title + "\n\n" + m["content"])

    return "\n".join(response)


def get_prompt_extraction_schema(
    original_prompt: str,
):
    required_variables = set(re.findall(r"\{(.+?)\}", original_prompt, re.MULTILINE))
    if required_variables:
        variables_str = ", ".join(f"{{{var}}}" for var in required_variables)
        prompt_description = (
            f" The prompt section being optimized contains the following f-string variables to be templated in: {variables_str}."
            " You must retain all of these variables in your improved prompt. No other input variables are allowed."
        )
    else:
        prompt_description = (
            " The prompt section being optimized contains no input f-string variables."
            " Any brackets {{ foo }} you emit will be escaped and not used."
        )

    pipeline = utils.get_var_healer(set(required_variables), all_required=True)

    class OptimizedPromptOutput(BaseModel):
        """Schema for the optimized prompt output."""

        analysis: str = Field(
            description="First, analyze the current results and plan improvements to reconcile them."
        )
        improved_prompt: typing.Optional[str] = Field(
            description="Finally, generate the full updated prompt to address the identified issues. "
            f" {prompt_description}"
        )

        @model_validator(mode="before")
        @classmethod
        def validate_input_variables(cls, data: typing.Any) -> typing.Any:
            assert "improved_prompt" in data
            data["improved_prompt"] = pipeline(data["improved_prompt"])
            return data

    return OptimizedPromptOutput



import asyncio
import typing
import uuid

from langchain.chat_models import init_chat_model
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import AIMessage, AnyMessage
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable, RunnableConfig, RunnableLambda
from langchain_core.runnables.config import get_executor_for_config
from langgraph.store.base import BaseStore, SearchItem
from langgraph.utils.config import get_store
from pydantic import BaseModel, Field
from trustcall import create_extractor
from typing_extensions import TypedDict

from langmem import utils
from langmem.knowledge.tools import create_search_memory_tool

## LangGraph Tools


class MessagesState(TypedDict):
    messages: list[AnyMessage]


class MemoryState(MessagesState):
    existing: typing.NotRequired[list[tuple[str, BaseModel]]]
    max_steps: int  # Default of 1


class SummarizeThread(BaseModel):
    title: str
    summary: str


class ExtractedMemory(typing.NamedTuple):
    id: str
    content: BaseModel


S = typing.TypeVar("S", bound=BaseModel)


class Memory(BaseModel):
    """Call this tool once for each new memory you want to record. Use multi-tool calling to record multiple new memories."""

    content: str = Field(
        description="The memory as a well-written, standalone episode/fact/note/preference/etc."
        " Refer to the user's instructions for more information the prefered memory organization."
    )


@typing.overload
def create_thread_extractor(
    model: str,
    /,
    schema: None = None,
    instructions: str = "You are tasked with summarizing the following conversation.",
) -> Runnable[MessagesState, SummarizeThread]: ...


@typing.overload
def create_thread_extractor(
    model: str,
    /,
    schema: S,
    instructions: str = "You are tasked with summarizing the following conversation.",
) -> Runnable[MessagesState, S]: ...


def create_thread_extractor(
    model: str,
    /,
    schema: typing.Union[None, BaseModel, type] = None,
    instructions: str = "You are tasked with summarizing the following conversation.",
) -> Runnable[MessagesState, BaseModel]:
    """Creates a conversation thread summarizer using schema-based extraction.

    This function creates an asynchronous callable that takes conversation messages and produces
    a structured summary based on the provided schema. If no schema is provided, it uses a default
    schema with title and summary fields.

    Args:
        model (str): The chat model to use for summarization (name or instance)
        schema (Optional[Union[BaseModel, type]], optional): Pydantic model for structured output.
            Defaults to a simple summary schema with title and summary fields.
        instructions (str, optional): System prompt template for the summarization task.
            Defaults to a basic summarization instruction.

    Returns:
        extractor (Callable[[list], typing.Awaitable[typing.Any]]): Async callable that takes a list of messages and returns a structured summary

    ???+ example "Examples"
        ```python
        from langmem import create_thread_extractor

        summarizer = create_thread_extractor("gpt-4")

        messages = [
            {"role": "user", "content": "Hi, I'm having trouble with my account"},
            {
                "role": "assistant",
                "content": "I'd be happy to help. What seems to be the issue?",
            },
            {"role": "user", "content": "I can't reset my password"},
        ]

        summary = await summarizer.ainvoke({"messages": messages})
        print(summary.title)
        # Output: "Password Reset Assistance"
        print(summary.summary)
        # Output: "User reported issues with password reset process..."
        ```

    """
    if schema is None:
        schema = SummarizeThread

    extractor = create_extractor(model, tools=[schema], tool_choice="any")

    template = ChatPromptTemplate.from_messages(
        [
            ("system", instructions),
            (
                "user",
                "Call the provided tool based on the conversation below:\n\n<conversation>{conversation}</conversation>",
            ),
        ]
    )

    def merge_messages(input: dict) -> dict:
        conversation = utils.get_conversation(input["messages"])

        return {"conversation": conversation} | {
            k: v for k, v in input.items() if k != "messages"
        }

    return (
        merge_messages | template | extractor | (lambda out: out["responses"][0])
    ).with_config({"run_name": "thread_extractor"})  # type: ignore


_MEMORY_INSTRUCTIONS = """You are a long-term memory manager maintaining a core store of semantic, procedural, and episodic memory. These memories power a life-long learning agent's core predictive model.

What should the agent learn from this interaction about the user, itself, or how it should act? Reflect on the input trajectory and current memories (if any).

1. **Extract & Contextualize**  
   - Identify essential facts, relationships, preferences, reasoning procedures, and context
   - Caveat uncertain or suppositional information with confidence levels (p(x)) and reasoning
   - Quote supporting information when necessary

2. **Compare & Update**  
   - Attend to novel information that deviates from existing memories and expectations.
   - Consolidate and compress redundant memories to maintain information-density; strengthen based on reliability and recency; maximize SNR by avoiding idle words.
   - Remove incorrect or redundant memories while maintaining internal consistency

3. **Synthesize & Reason**  
   - What can you conclude about the user, agent ("I"), or environment using deduction, induction, and abduction?
   - What patterns, relationships, and principles emerge about optimal responses?
   - What generalizations can you make?
   - Qualify conclusions with probabilistic confidence and justification

As the agent, record memory content exactly as you'd want to recall it when predicting how to act or respond. 
Prioritize retention of surprising (pattern deviation) and persistent (frequently reinforced) information, ensuring nothing worth remembering is forgotten and nothing false is remembered. Prefer dense, complete memories over overlapping ones."""


class Done(BaseModel):
    """Only call this tool once you are done forming & consolidating memories.
    Before that, continue to refine existing memories by patching and removing them
    or create new ones."""

    pass


class MemoryManager(Runnable[MemoryState, list[ExtractedMemory]]):
    def __init__(
        self,
        model: str | BaseChatModel,
        *,
        schemas: typing.Sequence[typing.Union[BaseModel, type]] = (Memory,),
        instructions: str = _MEMORY_INSTRUCTIONS,
        enable_inserts: bool = True,
        enable_updates: bool = True,
        enable_deletes: bool = False,
    ):
        self.model = (
            model if isinstance(model, BaseChatModel) else init_chat_model(model)
        )
        self.schemas = schemas or (Memory,)
        self.instructions = instructions
        self.enable_inserts = enable_inserts
        self.enable_updates = enable_updates
        self.enable_deletes = enable_deletes

    def _prepare_messages(
        self, messages: list[AnyMessage], max_steps: int = 1
    ) -> list[dict]:
        id_ = str(uuid.uuid4())
        session = (
            f"\n\n<session_{id_}>\n{utils.get_conversation(messages)}\n</session_{id_}>"
        )
        if max_steps > 1:
            session = f"{session}\n\nYou have a maximum of {max_steps - 1} attempts"
            " to form and consolidate memories from this session."
        return [
            {"role": "system", "content": "You are a memory subroutine for an AI."},
            {
                "role": "user",
                "content": (
                    f"{self.instructions}\n\nEnrich, prune, and organize memories based on any new information. "
                    f"If an existing memory is incorrect or outdated, update it based on the new information. "
                    f"All operations must be done in single parallel multi-tool call."
                    f" Avoid duplicate extractions. {session}"
                ),
            },
        ]

    def _prepare_existing(
        self,
        existing: typing.Optional[
            typing.Union[
                list[str], list[tuple[str, BaseModel]], list[tuple[str, str, dict]]
            ]
        ],
    ) -> list[tuple[str, str, typing.Any]]:
        if existing is None:
            return []
        if all(isinstance(ex, str) for ex in existing):
            MemoryModel = self.schemas[0]
            return [
                (str(uuid.uuid4()), "Memory", MemoryModel(content=ex))
                for ex in existing
            ]
        result = []
        for e in existing:
            if isinstance(e, (tuple, list)) and len(e) == 3:
                result.append(tuple(e))
            else:
                # Assume a two-element tuple: (id, value)
                id_, value = e[0], e[1]
                kind = (
                    value.__repr_name__() if isinstance(value, BaseModel) else "__any__"
                )
                result.append((id_, kind, value))
        return result

    @staticmethod
    def _filter_response(
        memories: list[ExtractedMemory],
        external_ids: set[str],
        exclude_removals: bool = False,
    ) -> list[ExtractedMemory]:
        """
        When exclude_removals is True (for the next iteration payload),
        drop any memory whose content is a RemoveDoc.
        When False (final response), drop removal objects only for internal memories.
        """
        results = []
        for rid, value in memories:
            is_removal = (
                hasattr(value, "__repr_name__") and value.__repr_name__() == "RemoveDoc"
            )
            if exclude_removals:
                if is_removal:
                    continue
            else:
                # Final response: if this is a removal *and* its id is not external, skip it.
                if is_removal and (rid not in external_ids):
                    continue
            results.append(ExtractedMemory(id=rid, content=value))
        return results

    async def ainvoke(
        self,
        input: MemoryState,
        config: typing.Optional[RunnableConfig] = None,
        **kwargs: typing.Any,
    ) -> list[ExtractedMemory]:
        max_steps = input.get("max_steps")
        if max_steps is None:
            max_steps = 1
        messages = input["messages"]
        existing = input.get("existing")
        prepared_messages = self._prepare_messages(messages, max_steps)
        prepared_existing = self._prepare_existing(existing)
        # Track external memory IDs (those passed in from outside)
        external_ids = {mem_id for mem_id, _, _ in prepared_existing}

        extractor = create_extractor(
            self.model,
            tools=list(self.schemas),
            enable_inserts=self.enable_inserts,
            enable_updates=self.enable_updates,
            enable_deletes=self.enable_deletes,
            existing_schema_policy=False,
        )
        # initial payload uses the full prepared_existing list
        payload = {"messages": prepared_messages, "existing": prepared_existing}
        # Use a dict to record the latest update for each memory id.
        results: dict[str, BaseModel] = {}

        for i in range(max_steps):
            if i == 1:
                extractor = create_extractor(
                    self.model,
                    tools=list(self.schemas) + [Done],
                    enable_inserts=self.enable_inserts,
                    enable_updates=self.enable_updates,
                    enable_deletes=self.enable_deletes,
                    existing_schema_policy=False,
                )
            response = await extractor.ainvoke(payload)
            is_done = False
            step_results = {}
            for r, rmeta in zip(response["responses"], response["response_metadata"]):
                if hasattr(r, "__repr_name__") and r.__repr_name__() == "Done":
                    is_done = True
                    continue
                mem_id = (
                    r.json_doc_id
                    if hasattr(r, "__repr_name__") and r.__repr_name__() == "RemoveDoc"
                    else rmeta.get("json_doc_id", str(uuid.uuid4()))
                )
                step_results[mem_id] = r
            results.update(step_results)

            for mem_id, _, mem in prepared_existing:
                if mem_id not in results:
                    results[mem_id] = mem

            ai_msg = response["messages"][-1]
            if is_done or not ai_msg.tool_calls:
                break
            if i < max_steps - 1:
                actions = [
                    (
                        "updated"
                        if rmeta.get("json_doc_id")
                        else (
                            "deleted"
                            if hasattr(r, "__repr_name__")
                            and r.__repr_name__() == "RemoveDoc"
                            else "inserted"
                        )
                    )
                    for r, rmeta in zip(
                        response["responses"], response["response_metadata"]
                    )
                ]
                prepared_messages = (
                    prepared_messages
                    + [response["messages"][-1]]
                    + [
                        {
                            "role": "tool",
                            "content": f"Memory {rid} {action}.",
                            "tool_call_id": tc["id"],
                        }
                        for tc, ((rid, _), action) in zip(
                            ai_msg.tool_calls, zip(list(step_results.items()), actions)
                        )
                    ]
                )
                # For the next iteration payload, drop all removal objects.
                payload = {
                    "messages": prepared_messages,
                    "existing": self._filter_response(
                        list(results.items()), external_ids, exclude_removals=True
                    ),
                }

        # For the final response, include removals only if they refer to an external memory.
        return self._filter_response(
            list(results.items()), external_ids, exclude_removals=False
        )

    def invoke(
        self,
        input: MemoryState,
        config: typing.Optional[RunnableConfig] = None,
        **kwargs: typing.Any,
    ) -> list[ExtractedMemory]:
        max_steps = input.get("max_steps")
        if max_steps is None:
            max_steps = 1
        messages = input["messages"]
        existing = input.get("existing")
        prepared_messages = self._prepare_messages(messages, max_steps)
        prepared_existing = self._prepare_existing(existing)
        # Track external memory IDs (those passed in from outside)
        external_ids = {mem_id for mem_id, _, _ in prepared_existing}

        extractor = create_extractor(
            self.model,
            tools=list(self.schemas),
            enable_inserts=self.enable_inserts,
            enable_updates=self.enable_updates,
            enable_deletes=self.enable_deletes,
            existing_schema_policy=False,
        )
        payload = {"messages": prepared_messages, "existing": prepared_existing}
        # Use a dict to record the latest update for each memory id.
        results: dict[str, BaseModel] = {}

        for i in range(max_steps):
            if i == 1:
                extractor = create_extractor(
                    self.model,
                    tools=list(self.schemas) + [Done],
                    enable_inserts=self.enable_inserts,
                    enable_updates=self.enable_updates,
                    enable_deletes=self.enable_deletes,
                    existing_schema_policy=False,
                )
            response = extractor.invoke(payload)
            is_done = False
            step_results: dict[str, BaseModel] = {}
            for r, rmeta in zip(response["responses"], response["response_metadata"]):
                if hasattr(r, "__repr_name__") and r.__repr_name__() == "Done":
                    is_done = True
                    continue
                mem_id = (
                    r.json_doc_id
                    if (
                        hasattr(r, "__repr_name__") and r.__repr_name__() == "RemoveDoc"
                    )
                    else rmeta.get("json_doc_id", str(uuid.uuid4()))
                )
                step_results[mem_id] = r
            results.update(step_results)

            # Ensure any memory from the initial payload that hasn't been updated is retained.
            for mem_id, _, mem in prepared_existing:
                if mem_id not in results:
                    results[mem_id] = mem

            ai_msg = response["messages"][-1]
            if is_done or not ai_msg.tool_calls:
                break
            if i < max_steps - 1:
                actions = [
                    (
                        "updated"
                        if rmeta.get("json_doc_id")
                        else (
                            "deleted"
                            if (
                                hasattr(r, "__repr_name__")
                                and r.__repr_name__() == "RemoveDoc"
                            )
                            else "inserted"
                        )
                    )
                    for r, rmeta in zip(
                        response["responses"], response["response_metadata"]
                    )
                ]
                prepared_messages = (
                    prepared_messages
                    + [response["messages"][-1]]
                    + [
                        {
                            "role": "tool",
                            "content": f"Memory {rid} {action}.",
                            "tool_call_id": tc["id"],
                        }
                        for tc, ((rid, _), action) in zip(
                            ai_msg.tool_calls, zip(list(step_results.items()), actions)
                        )
                    ]
                )
                payload = {
                    "messages": prepared_messages,
                    "existing": self._filter_response(
                        list(results.items()), external_ids, exclude_removals=True
                    ),
                }

        return self._filter_response(
            list(results.items()), external_ids, exclude_removals=False
        )

    async def __call__(
        self,
        messages: typing.Sequence[AnyMessage],
        existing: typing.Optional[typing.Sequence[ExtractedMemory]] = None,
    ) -> list[ExtractedMemory]:
        input: MemoryState = {"messages": messages}
        if existing is not None:
            input["existing"] = existing
        return await self.ainvoke(input)


def create_memory_manager(
    model: str | BaseChatModel,
    /,
    *,
    schemas: typing.Sequence[typing.Union[BaseModel, type]] = (Memory,),
    instructions: str = _MEMORY_INSTRUCTIONS,
    enable_inserts: bool = True,
    enable_updates: bool = True,
    enable_deletes: bool = False,
) -> Runnable[MemoryState, list[ExtractedMemory]]:
    """Create a memory manager that processes conversation messages and generates structured memory entries.

    This function creates an async callable that analyzes conversation messages and existing memories
    to generate or update structured memory entries. It can identify implicit preferences,
    important context, and key information from conversations, organizing them into
    well-structured memories that can be used to improve future interactions.

    The manager supports both unstructured string-based memories and structured memories
    defined by Pydantic models, all automatically persisted to the configured storage.

    Args:
        model (Union[str, BaseChatModel]): The language model to use for memory enrichment.
            Can be a model name string or a BaseChatModel instance.
        schemas (Optional[list]): List of Pydantic models defining the structure of memory
            entries. Each model should define the fields and validation rules for a type
            of memory. If None, uses unstructured string-based memories. Defaults to None.
        instructions (str, optional): Custom instructions for memory generation and
            organization. These guide how the model extracts and structures information
            from conversations. Defaults to predefined memory instructions.
        enable_inserts (bool, optional): Whether to allow creating new memory entries.
            When False, the manager will only update existing memories. Defaults to True.
        enable_updates (bool, optional): Whether to allow updating existing memories
            that are outdated or contradicted by new information. Defaults to True.
        enable_deletes (bool, optional): Whether to allow deleting existing memories
            that are outdated or contradicted by new information. Defaults to False.

    Returns:
        manager: An runnable that processes conversations and returns `ExtractedMemory`'s. The function signature depends on whether schemas are provided

    ???+ example "Examples"
        Basic unstructured memory enrichment:
        ```python
        from langmem import create_memory_manager

        manager = create_memory_manager("anthropic:claude-3-5-sonnet-latest")

        conversation = [
            {"role": "user", "content": "I prefer dark mode in all my apps"},
            {"role": "assistant", "content": "I'll remember that preference"},
        ]

        # Extract memories from conversation
        memories = await manager(conversation)
        print(memories[0][1])  # First memory's content
        # Output: "User prefers dark mode for all applications"
        ```

        Structured memory enrichment with Pydantic models:
        ```python
        from pydantic import BaseModel
        from langmem import create_memory_manager

        class PreferenceMemory(BaseModel):
            \"\"\"Store the user's preference\"\"\"
            category: str
            preference: str
            context: str

        manager = create_memory_manager(
            "anthropic:claude-3-5-sonnet-latest",
            schemas=[PreferenceMemory]
        )

        # Same conversation, but with structured output
        conversation = [
            {"role": "user", "content": "I prefer dark mode in all my apps"},
            {"role": "assistant", "content": "I'll remember that preference"}
        ]
        memories = await manager(conversation)
        print(memories[0][1])
        # Output:
        # PreferenceMemory(
        #     category="ui",
        #     preference="dark_mode",
        #     context="User explicitly stated preference for dark mode in all applications"
        # )
        ```

        Working with existing memories:
        ```python
        conversation = [
            {
                "role": "user",
                "content": "Actually I changed my mind, dark mode hurts my eyes",
            },
            {"role": "assistant", "content": "I'll update your preference"},
        ]

        # The manager will upsert; working with the existing memory instead of always creating a new one
        updated_memories = await manager.ainvoke(
            {"messages": conversation, "existing": memories}
        )
        ```

        Insertion-only memories:
        ```python
        manager = create_memory_manager(
            "anthropic:claude-3-5-sonnet-latest",
            schemas=[PreferenceMemory],
            enable_updates=False,
            enable_deletes=False,
        )

        conversation = [
            {
                "role": "user",
                "content": "Actually I changed my mind, dark mode is the best mode",
            },
            {"role": "assistant", "content": "I'll update your preference"},
        ]

        # The manager will only create new memories
        updated_memories = await manager.ainvoke(
            {"messages": conversation, "existing": memories}
        )
        print(updated_memories)
        ```

        Providing multiple max steps for extraction and synthesis:
        ```python
        manager = create_memory_manager(
            "anthropic:claude-3-5-sonnet-latest",
            schemas=[PreferenceMemory],
        )

        conversation = [
            {"role": "user", "content": "I prefer dark mode in all my apps"},
            {"role": "assistant", "content": "I'll remember that preference"},
        ]

        # Set max steps for extraction and synthesis
        max_steps = 3
        memories = await manager.ainvoke({"messages": conversation, "max_steps": max_steps})
        print(memories)
        ```
    """

    return MemoryManager(
        model,
        schemas=schemas,
        instructions=instructions,
        enable_inserts=enable_inserts,
        enable_updates=enable_updates,
        enable_deletes=enable_deletes,
    )


def create_memory_searcher(
    model: str | BaseChatModel,
    prompt: str = "Search for distinct memories relevant to different aspects of the provided context.",
    *,
    namespace: tuple[str, ...] = ("memories", "{langgraph_user_id}"),
) -> Runnable[MessagesState, typing.Awaitable[list[SearchItem]]]:
    """Creates a memory search pipeline with automatic query generation.

    This function builds a pipeline that combines query generation, memory search,
    and result ranking into a single component. It uses the provided model to
    generate effective search queries based on conversation context.

    Args:
        model (Union[str, BaseChatModel]): The language model to use for search query generation.
            Can be a model name string or a BaseChatModel instance.
        prompt (str, optional): System prompt template for search assistant.
            Defaults to a basic search prompt.
        namespace: The namespace structure for organizing memories in LangGraph's BaseStore.
            Uses runtime configuration with placeholders like `{langgraph_user_id}`.
            See [Memory Namespaces](../concepts/conceptual_guide.md#memory-namespaces).
            Defaults to ("memories", "{langgraph_user_id}").

    ???+ note "Namespace Configuration"
        If the namespae has template variables "{variable_name}", they will be configured at
        runtime through the `config` parameter:
        ```python
        # Example: Search user's memories
        config = {"configurable": {"langgraph_user_id": "user-123"}}
        # Searches in namespace: ("memories", "user-123")

        # Example: Search team knowledge
        config = {"configurable": {"langgraph_user_id": "team-x"}}
        # Searches in namespace: ("memories", "team-x")
        ```

    Returns:
        searcher (Callable[[list], typing.Awaitable[typing.Any]]): A pipeline that takes conversation messages and returns sorted memory artifacts,
            ranked by relevance score.

    ???+ example "Examples"
        ```python
        from langmem import create_memory_searcher
        from langgraph.store.memory import InMemoryStore
        from langgraph.func import entrypoint

        store = InMemoryStore(
            index={
                "dims": 1536,
                "embed": "openai:text-embedding-3-small",
            }
        )
        user_id = "abcd1234"
        store.put(
            ("memories", user_id), key="preferences", value={"content": "I like sushi"}
        )
        searcher = create_memory_searcher(
            "openai:gpt-4o-mini", namespace=("memories", "{langgraph_user_id}")
        )


        @entrypoint(store=store)
        async def search_memories(messages: list):
            results = await searcher.ainvoke({"messages": messages})
            print(results[0].value["content"])
            # Output: "I like sushi"


        await search_memories.ainvoke(
            [{"role": "user", "content": "What do I like to eat?"}],
            config={"configurable": {"langgraph_user_id": user_id}},
        )
        ```

    """
    template = ChatPromptTemplate.from_messages(
        [
            ("system", prompt),
            ("placeholder", "{messages}"),
            ("user", "\n\nSearch for memories relevant to the above context."),
        ]
    )

    # Initialize model and search tool
    model_instance = (
        model if isinstance(model, BaseChatModel) else init_chat_model(model)
    )
    search_tool = create_search_memory_tool(
        namespace=namespace, response_format="content_and_artifact"
    )
    query_gen = model_instance.bind_tools([search_tool], tool_choice="search_memory")

    def return_sorted(tool_messages: list):
        artifacts = {
            (*item.namespace, item.key): item
            for msg in tool_messages
            for item in (msg.artifact or [])
        }
        return [
            v
            for v in sorted(
                artifacts.values(),
                key=lambda item: item.score if item.score is not None else 0,
                reverse=True,
            )
        ]

    def search(msg: AIMessage) -> list[SearchItem]:
        return search_tool.batch(
            [tc for tc in msg.tool_calls if tc["name"] == "search_memory"]
        )

    async def search_async(msg: AIMessage) -> list[SearchItem]:
        return await search_tool.abatch(
            [tc for tc in msg.tool_calls if tc["name"] == "search_memory"]
        )

    searcher = RunnableLambda(search, search_async)

    return (  # type: ignore
        template | utils.merge_message_runs | query_gen | searcher | return_sorted
    ).with_config({"run_name": "search_memory_pipeline"})


class MemoryPhase(TypedDict, total=False):
    instructions: str
    include_messages: bool
    enable_inserts: bool
    enable_deletes: bool


class MemoryStoreManagerInput(TypedDict):
    """Input schema for MemoryStoreManager."""

    messages: list[AnyMessage]
    max_steps: int  # Default of 1


class MemoryStoreManager(Runnable[MemoryStoreManagerInput, list[dict]]):
    def __init__(
        self,
        model: str | BaseChatModel,
        /,
        *,
        schemas: list | None = None,
        instructions: str = _MEMORY_INSTRUCTIONS,
        enable_inserts: bool = True,
        enable_deletes: bool = True,
        query_model: str | BaseChatModel | None = None,
        query_limit: int = 5,
        namespace: tuple[str, ...] = ("memories", "{langgraph_user_id}"),
        store: BaseStore | None = None,
        phases: list[MemoryPhase] | None = None,
    ):
        self.model = (
            model if isinstance(model, BaseChatModel) else init_chat_model(model)
        )
        self.query_model = (
            None
            if query_model is None
            else (
                query_model
                if isinstance(query_model, BaseChatModel)
                else init_chat_model(query_model)
            )
        )
        self.schemas = schemas
        self.instructions = instructions
        self.enable_inserts = enable_inserts
        self.enable_deletes = enable_deletes
        self.query_limit = query_limit
        self.phases = phases or []
        self.namespace = utils.NamespaceTemplate(namespace)
        self.store = store

        self.memory_manager = create_memory_manager(
            self.model,
            schemas=schemas,
            instructions=instructions,
            enable_inserts=enable_inserts,
            enable_deletes=enable_deletes,
        )
        self.search_tool = create_search_memory_tool(
            namespace=namespace,
            instructions="Queries should be formatted as hypothetical memories that would be relevant to the current conversation.",
        )
        self.query_gen = None
        if self.query_model is not None:
            self.query_gen = self.query_model.bind_tools(
                [self.search_tool], tool_choice="any"
            )

    def _get_store(self) -> BaseStore:
        """Get the store to use for memory storage.

        Returns the store provided during initialization if available,
        otherwise falls back to the store configured in the LangGraph context.
        """
        if self.store is not None:
            return self.store
        return get_store()

    @staticmethod
    def _stable_id(item: SearchItem) -> str:
        return uuid.uuid5(uuid.NAMESPACE_DNS, str((*item.namespace, item.key))).hex

    @staticmethod
    def _apply_manager_output(
        manager_output: list[ExtractedMemory],
        store_based: list[tuple[str, str, dict]],
        store_map: dict[str, SearchItem],
        ephemeral: list[tuple[str, str, dict]],
    ) -> tuple[list[tuple[str, str, dict]], list[tuple[str, str, dict]], list[str]]:
        store_dict = {sid: (sid, kind, content) for (sid, kind, content) in store_based}
        ephemeral_dict = {
            sid: (sid, kind, content) for (sid, kind, content) in ephemeral
        }
        removed_ids = []
        for extracted in manager_output:
            stable_id = extracted.id
            model_data = extracted.content
            if isinstance(model_data, BaseModel):
                if (
                    hasattr(model_data, "__repr_name__")
                    and model_data.__repr_name__() == "RemoveDoc"
                ):
                    removal_id = getattr(model_data, "json_doc_id", None)
                    if removal_id and removal_id in store_map:
                        removed_ids.append(removal_id)
                    store_dict.pop(removal_id, None)
                    ephemeral_dict.pop(removal_id, None)
                    continue
                new_content = model_data.model_dump(mode="json")
                new_kind = model_data.__repr_name__()
            else:
                new_kind = store_dict.get(stable_id, (stable_id, "Memory", {}))[1]
                new_content = model_data
            if stable_id in store_dict:
                store_dict[stable_id] = (stable_id, new_kind, new_content)
            else:
                ephemeral_dict[stable_id] = (stable_id, new_kind, new_content)
        return list(store_dict.values()), list(ephemeral_dict.values()), removed_ids

    def _build_phase_manager(
        self, phase: MemoryPhase
    ) -> Runnable[MessagesState, list[ExtractedMemory]]:
        return create_memory_manager(
            self.model,
            schemas=self.schemas,
            instructions=phase.get(
                "instructions",
                "You are a memory manager. Deduplicate, consolidate, and enrich these memories.",
            ),
            enable_inserts=phase.get("enable_inserts", True),
            enable_deletes=phase.get("enable_deletes", True),
        )

    @staticmethod
    def _sort_results(
        search_results_lists: list[list[SearchItem]], query_limit: int
    ) -> dict[str, SearchItem]:
        search_results = {}
        for results in search_results_lists:
            for item in results:
                search_results[(tuple(item.namespace), item.key)] = item
        sorted_results = sorted(
            search_results.values(),
            key=lambda it: it.score if it.score is not None else float("-inf"),
            reverse=True,
        )[:query_limit]
        return {MemoryStoreManager._stable_id(item): item for item in sorted_results}

    async def ainvoke(
        self,
        input: MemoryStoreManagerInput,
        config: typing.Optional[RunnableConfig] = None,
        **kwargs: typing.Any,
    ) -> list[dict]:
        store = self._get_store()
        namespace = self.namespace(config)

        if self.query_gen:
            convo = utils.get_conversation(input["messages"])
            query_text = (
                f"Use parallel tool calling to search for distinct memories relevant to this conversation.:\n\n"
                f"<convo>\n{convo}\n</convo>."
            )
            query_req = await self.query_gen.ainvoke(query_text)
            search_results_lists = await asyncio.gather(
                *[
                    store.asearch(
                        namespace, **({**tc["args"], "limit": self.query_limit})
                    )
                    for tc in query_req.tool_calls
                ]
            )
        else:
            # Search over "query_limit" timespans starting from the most recent
            queries = utils.get_dialated_windows(
                input["messages"], self.query_limit // 4
            )
            search_results_lists = await asyncio.gather(
                *[store.asearch(namespace, query=query) for query in queries]
            )

        store_map = self._sort_results(search_results_lists, self.query_limit)

        store_based = [
            (sid, item.value["kind"], item.value["content"])
            for sid, item in store_map.items()
        ]
        ephemeral: list[tuple[str, str, dict]] = []
        removed_ids: set[str] = set()

        # --- Enrich memories using the composed MemoryManager (async) ---
        enriched = await self.memory_manager.ainvoke(
            {
                "messages": input["messages"],
                "existing": store_based,
                "max_steps": input.get("max_steps"),
            }
        )
        store_based, ephemeral, removed = self._apply_manager_output(
            enriched, store_based, store_map, ephemeral
        )
        removed_ids.update(removed)

        # Process additional phases.
        for phase in self.phases:
            phase_manager = self._build_phase_manager(phase)
            phase_messages = (
                input["messages"] if phase.get("include_messages", False) else []
            )
            phase_input = {
                "messages": phase_messages,
                "existing": store_based + ephemeral,
            }
            phase_enriched = await phase_manager.ainvoke(phase_input)
            store_based, ephemeral, removed = self._apply_manager_output(
                phase_enriched, store_based, store_map, ephemeral
            )
            removed_ids.update(removed)

        final_mem = store_based + ephemeral
        final_puts = []
        for sid, kind, content in final_mem:
            if sid in removed_ids:
                continue
            if sid in store_map:
                old_art = store_map[sid]
                if old_art.value["kind"] != kind or old_art.value["content"] != content:
                    final_puts.append(
                        {
                            "namespace": old_art.namespace,
                            "key": old_art.key,
                            "value": {"kind": kind, "content": content},
                        }
                    )
            else:
                final_puts.append(
                    {
                        "namespace": namespace,
                        "key": sid,
                        "value": {"kind": kind, "content": content},
                    }
                )

        final_deletes = []
        for sid in removed_ids:
            if sid in store_map:
                art = store_map[sid]
                final_deletes.append((art.namespace, art.key))

        await asyncio.gather(
            *(store.aput(**put) for put in final_puts),
            *(store.adelete(ns, key) for (ns, key) in final_deletes),
        )

        return final_puts

    def invoke(
        self,
        input: MemoryStoreManagerInput,
        config: typing.Optional[RunnableConfig] = None,
        **kwargs: typing.Any,
    ) -> list[dict]:
        store = self._get_store()
        namespace = self.namespace(config)
        convo = utils.get_conversation(input["messages"])

        with get_executor_for_config(config) as executor:
            if self.query_gen:
                convo = utils.get_conversation(input["messages"])
                query_text = (
                    f"Use parallel tool calling to search for distinct memories relevant to this conversation.:\n\n"
                    f"<convo>\n{convo}\n</convo>."
                )
                query_req = self.query_gen.invoke(query_text)
                search_results_futs = [
                    executor.submit(
                        store.search,
                        namespace,
                        **({**tc["args"], "limit": self.query_limit}),
                    )
                    for tc in query_req.tool_calls
                ]
            else:
                # Search over "query_limit" timespans starting from the most recent
                queries = utils.get_dialated_windows(
                    input["messages"], self.query_limit // 4
                )
                search_results_lists = [
                    store.search(namespace, query=query) for query in queries
                ]
                search_results_futs = [
                    executor.submit(
                        store.search,
                        namespace,
                        query=query,
                        limit=self.query_limit,
                    )
                    for query in queries
                ]

        search_results_lists = [fut.result() for fut in search_results_futs]
        store_map = self._sort_results(search_results_lists, self.query_limit)
        store_based = [
            (sid, item.value["kind"], item.value["content"])
            for sid, item in store_map.items()
        ]
        ephemeral: list[tuple[str, str, dict]] = []
        removed_ids: set[str] = set()

        enriched = self.memory_manager.invoke(
            {
                "messages": input["messages"],
                "existing": store_based,
                "max_steps": input.get("max_steps"),
            }
        )
        store_based, ephemeral, removed = self._apply_manager_output(
            enriched, store_based, store_map, ephemeral
        )
        removed_ids.update(removed)

        for phase in self.phases:
            phase_manager = self._build_phase_manager(phase)
            phase_messages = (
                input["messages"] if phase.get("include_messages", False) else []
            )
            phase_input = {
                "messages": phase_messages,
                "existing": store_based + ephemeral,
            }
            phase_enriched = phase_manager.invoke(phase_input)
            store_based, ephemeral, removed = self._apply_manager_output(
                phase_enriched, store_based, store_map, ephemeral
            )
            removed_ids.update(removed)

        final_mem = store_based + ephemeral
        final_puts = []
        for sid, kind, content in final_mem:
            if sid in removed_ids:
                continue
            if sid in store_map:
                old_art = store_map[sid]
                if old_art.value["kind"] != kind or old_art.value["content"] != content:
                    final_puts.append(
                        {
                            "namespace": old_art.namespace,
                            "key": old_art.key,
                            "value": {"kind": kind, "content": content},
                        }
                    )
            else:
                final_puts.append(
                    {
                        "namespace": namespace,
                        "key": sid,
                        "value": {"kind": kind, "content": content},
                    }
                )

        final_deletes = []
        for sid in removed_ids:
            if sid in store_map:
                art = store_map[sid]
                final_deletes.append((art.namespace, art.key))

        with get_executor_for_config(config) as executor:
            for put in final_puts:
                executor.submit(store.put, **put)
            for ns, key in final_deletes:
                executor.submit(store.delete, ns, key)

        return final_puts

    async def __call__(self, messages: typing.Sequence[AnyMessage]) -> list[dict]:
        return await self.ainvoke({"messages": messages})


def create_memory_store_manager(
    model: str | BaseChatModel,
    /,
    *,
    schemas: list | None = None,
    instructions: str = _MEMORY_INSTRUCTIONS,
    enable_inserts: bool = True,
    enable_deletes: bool = True,
    query_model: str | BaseChatModel | None = None,
    query_limit: int = 5,
    namespace: tuple[str, ...] = ("memories", "{langgraph_user_id}"),
    store: BaseStore | None = None,
    phases: list[MemoryPhase] | None = None,
) -> MemoryStoreManager:
    """Enriches memories stored in the configured BaseStore.

    The system automatically searches for relevant memories, extracts new information,
    updates existing memories, and maintains a versioned history of all changes.

    Args:
        model (Union[str, BaseChatModel]): The primary language model to use for memory
            enrichment. Can be a model name string or a BaseChatModel instance.
        schemas (Optional[list]): List of Pydantic models defining the structure of memory
            entries. Each model should define the fields and validation rules for a type
            of memory. If None, uses unstructured string-based memories. Defaults to None.
        instructions (str, optional): Custom instructions for memory generation and
            organization. These guide how the model extracts and structures information
            from conversations. Defaults to predefined memory instructions.
        enable_inserts (bool, optional): Whether to allow creating new memory entries.
            When False, the manager will only update existing memories. Defaults to True.
        enable_deletes (bool, optional): Whether to allow deleting existing memories
            that are outdated or contradicted by new information. Defaults to True.
        query_model (Optional[Union[str, BaseChatModel]], optional): Optional separate
            model for memory search queries. Using a smaller, faster model here can
            improve performance. If None, uses the primary model. Defaults to None.
        query_limit (int, optional): Maximum number of relevant memories to retrieve
            for each conversation. Higher limits provide more context but may slow
            down processing. Defaults to 5.
        namespace (tuple[str, ...], optional): Storage namespace structure for
            organizing memories. Supports templated values like "{langgraph_user_id}" which are
            populated from the runtime context. Defaults to `("memories", "{langgraph_user_id}")`.
        store (Optional[BaseStore], optional): The store to use for memory storage.
            If None, uses the store configured in the LangGraph config. Defaults to None.
            When using LangGraph Platform, the server will manage the store for you.
        phases (Optional[list]): List of MemoryPhase objects defining the phases of the memory enrichment process.

    Returns:
        manager: An runnable that processes conversations and automatically manages memories in the LangGraph BaseStore.

    The basic data flow works as follows:

    ```mermaid
    sequenceDiagram
    participant Client
    participant Manager
    participant Store
    participant LLM

    Client->>Manager: conversation history
    Manager->>Store: find similar memories
    Store-->>Manager: memories
    Manager->>LLM: analyze & extract
    LLM-->>Manager: memory updates
    Manager->>Store: apply changes
    Manager-->>Client: updated memories
    ```

    ???+ example "Examples"
        Run memory extraction "inline" within your LangGraph app.
        By default, each "memory" is a simple string:
        ```python
        import os

        from anthropic import AsyncAnthropic
        from langchain_core.runnables import RunnableConfig
        from langgraph.func import entrypoint
        from langgraph.store.memory import InMemoryStore

        from langmem import create_memory_store_manager

        store = InMemoryStore(
            index={
                "dims": 1536,
                "embed": "openai:text-embedding-3-small",
            }
        )

        manager = create_memory_store_manager("anthropic:claude-3-5-sonnet-latest", namespace=("memories", "{langgraph_user_id}"))
        client = AsyncAnthropic(api_key=os.getenv("ANTHROPIC_API_KEY"))


        @entrypoint(store=store)
        async def my_agent(message: str, config: RunnableConfig):
            memories = await store.asearch(
                ("memories", config["configurable"]["langgraph_user_id"]),
                query=message,
            )
            llm_response = await client.messages.create(
                model="claude-3-5-sonnet-latest",
                system="You are a helpful assistant.\\n\\n## Memories from the user:"
                f"\\n<memories>\\n{memories}\\n</memories>",
                max_tokens=2048,
                messages=[{"role": "user", "content": message}],
            )
            response = {"role": "assistant", "content": llm_response.content[0].text}

            await manager.ainvoke(
                {"messages": [{"role": "user", "content": message}, response]},
            )
            return response["content"]


        response_1 = await my_agent.ainvoke(
            "I prefer dark mode in all my apps",
            config={"configurable": {"langgraph_user_id": "user123"}},
        )
        print("response_1:", response_1)
        # Later conversation - automatically retrieves and uses the stored preference
        response_2 = await my_agent.ainvoke(
            "What theme do I prefer?",
            config={"configurable": {"langgraph_user_id": "user123"}},
        )
        print("response_2:", response_2)
        # You can list over memories in the user's namespace manually:
        print(store.search(("memories", "user123")))
        ```

        You can customize what each memory can look like by defining **schemas**:
        ```python
        from langgraph.func import entrypoint
        from langgraph.store.memory import InMemoryStore
        from pydantic import BaseModel

        from langmem import create_memory_store_manager

        store = InMemoryStore(
            index={
                "dims": 1536,
                "embed": "openai:text-embedding-3-small",
            }
        )
        manager = create_memory_store_manager(
            "anthropic:claude-3-5-sonnet-latest",
            namespace=("memories", "{langgraph_user_id}"),
        )

        class PreferenceMemory(BaseModel):
            \"\"\"Store preferences about the user.\"\"\"
            category: str
            preference: str
            context: str


        store = InMemoryStore(
            index={
                "dims": 1536,
                "embed": "openai:text-embedding-3-small",
            }
        )
        manager = create_memory_store_manager(
            "anthropic:claude-3-5-sonnet-latest",
            schemas=[PreferenceMemory],
            namespace=("project", "team_1", "{langgraph_user_id}"),
        )


        @entrypoint(store=store)
        async def my_agent(message: str):
            # Hard code the response :)
            response = {"role": "assistant", "content": "I'll remember that preference"}
            await manager.ainvoke(
                {"messages": [{"role": "user", "content": message}, response]}
            )
            return response


        # Store structured memory
        await my_agent.ainvoke(
            "I prefer dark mode in all my apps",
            config={"configurable": {"langgraph_user_id": "user123"}},
        )

        # See the extracted memories yourself
        print(store.search(("memories", "user123")))

        # Memory is automatically stored and can be retrieved in future conversations
        # The system will also automatically update it if preferences change
        ```

    By default, relevant memories are recalled by directly embedding the new messages. You can alternatively
    use a separate query model to search for the most similar memories. Here's how it works:

    ```mermaid
        sequenceDiagram
            participant Client
            participant Manager
            participant QueryLLM
            participant Store
            participant MainLLM

            Client->>Manager: messages
            Manager->>QueryLLM: generate search query
            QueryLLM-->>Manager: optimized query
            Manager->>Store: find memories
            Store-->>Manager: memories
            Manager->>MainLLM: analyze & extract
            MainLLM-->>Manager: memory updates
            Manager->>Store: apply changes
            Manager-->>Client: result
    ```

    ???+ example "Using an LLM to search for memories"
        ```python
        from langmem import create_memory_store_manager
        from langgraph.store.memory import InMemoryStore
        from langgraph.func import entrypoint

        store = InMemoryStore(
            index={
                "dims": 1536,
                "embed": "openai:text-embedding-3-small",
            }
        )
        manager = create_memory_store_manager(
            "anthropic:claude-3-5-sonnet-latest",  # Main model for memory processing
            query_model="anthropic:claude-3-5-haiku-latest",  # Faster model for search
            query_limit=10,  # Retrieve more relevant memories
            namespace=("memories", "{langgraph_user_id}"),
        )


        @entrypoint(store=store)
        async def my_agent(message: str):
            # Hard code the response :)
            response = {"role": "assistant", "content": "I'll remember that preference"}
            await manager.ainvoke(
                {"messages": [{"role": "user", "content": message}, response]}
            )
            return response


        await my_agent.ainvoke(
            "I prefer dark mode in all my apps",
            config={"configurable": {"langgraph_user_id": "user123"}},
        )

        # See the extracted memories yourself
        print(store.search(("memories", "user123")))
        ```

    In the examples above, we were calling the manager in the main thread. In a real application, you'll
    likely want to background the execution of the manager, either by executing it in a background thread or on a separate server.
    To do so, you can use the `ReflectionExecutor` class:

    ```mermaid
    sequenceDiagram
        participant Agent
        participant Background
        participant Store

        Agent->>Agent: process message
        Agent-->>User: response
        Agent->>Background: schedule enrichment<br/>(after_seconds=0)
        Note over Background,Store: Memory processing happens<br/>in background thread
    ```

    ???+ example "Running reflections in the background"
        Background enrichment using @entrypoint:
        ```python
        from langmem import create_memory_store_manager, ReflectionExecutor
        from langgraph.prebuilt import create_react_agent
        from langgraph.store.memory import InMemoryStore
        from langgraph.func import entrypoint

        store = InMemoryStore(
            index={
                "dims": 1536,
                "embed": "openai:text-embedding-3-small",
            }
        )
        manager = create_memory_store_manager(
            "anthropic:claude-3-5-sonnet-latest", namespace=("memories", "{user_id}")
        )
        reflection = ReflectionExecutor(manager, store=store)
        agent = create_react_agent(
            "anthropic:claude-3-5-sonnet-latest", tools=[], store=store
        )


        @entrypoint(store=store)
        async def chat(messages: list):
            response = await agent.ainvoke({"messages": messages})

            fut = reflection.submit(
                {
                    "messages": response["messages"],
                },
                # We'll schedule this immediately.
                # Adding a delay lets you **debounce** and deduplicate reflection work
                # whenever the user is actively engaging with the agent.
                after_seconds=0,
            )

            return fut


        fut = await chat.ainvoke(
            [{"role": "user", "content": "I prefer dark mode in my apps"}],
            config={"configurable": {"user_id": "user-123"}},
        )
        # Inspect the result
        fut.result()  # Wait for the reflection to complete; This is only for demoing the search inline
        print(store.search(("memories", "user-123")))
        ```
    """
    return MemoryStoreManager(
        model,
        schemas=schemas,
        instructions=instructions,
        enable_inserts=enable_inserts,
        enable_deletes=enable_deletes,
        query_model=query_model,
        query_limit=query_limit,
        namespace=namespace,
        store=store,
        phases=phases,
    )


__all__ = [
    "create_memory_manager",
    "create_memory_searcher",
    "create_memory_store_manager",
    "create_thread_extractor",
]



"""Knowledge extraction and semantic memory management.

This module provides utilities for extracting and managing semantic knowledge from conversations:

1. Functional Transformations:
    - `create_memory_manager(model, schemas=None) -> ((messages, existing=None) -> list[tuple[str, Memory]])`:
        Extract structured information from conversations
    - `create_thread_extractor(model, schema=None) -> ((messages) -> Summary)`:
        Generate structured summaries from conversations

2. Stateful Operations:
    Components that persist and manage memories in LangGraph's BaseStore:
    - `create_memory_store_manager(model, store=None) -> ((messages) -> None)`:
        Apply enrichment with integrated storage
    - `create_manage_memory_tool(store=None) -> Tool[dict, str]`:
        Tool for creating/updating stored memories
    - `create_search_memory_tool(store=None) -> Tool[dict, list[Memory]]`:
        Tool for searching stored memories

"""

from langmem.knowledge.extraction import (
    MemoryPhase,
    create_memory_manager,
    create_memory_searcher,
    create_memory_store_manager,
    create_thread_extractor,
)
from langmem.knowledge.tools import create_manage_memory_tool, create_search_memory_tool

__all__ = [
    "create_manage_memory_tool",
    "create_search_memory_tool",
    "create_memory_manager",
    "create_memory_searcher",
    "create_memory_store_manager",
    "create_thread_extractor",
    "MemoryPhase",
]



import logging
import typing
import uuid

from langchain_core.tools import StructuredTool
from langgraph.store.base import BaseStore
from langgraph.utils.config import get_store

from langmem import errors, utils

logger = logging.getLogger(__name__)

# LangGraph Tools


def create_manage_memory_tool(
    namespace: tuple[str, ...] | str,
    *,
    instructions: str = "Proactively call this tool when you:\n\n"
    "1. Identify a new USER preference.\n"
    "2. Receive an explicit USER request to remember something or otherwise alter your behavior.\n"
    "3. Are working and want to record important context.\n"
    "4. Identify that an existing MEMORY is incorrect or outdated.\n",
    schema: typing.Type = str,
    actions_permitted: typing.Optional[
        tuple[typing.Literal["create", "update", "delete"], ...]
    ] = ("create", "update", "delete"),
    store: typing.Optional[BaseStore] = None,
    name: str = "manage_memory",
):
    """Create a tool for managing persistent memories in conversations.

    This function creates a tool that allows AI assistants to create, update, and delete
    persistent memories that carry over between conversations. The tool helps maintain
    context and user preferences across sessions.


    Args:
        instructions: Custom instructions for when to use the memory tool.
            Defaults to a predefined set of guidelines for proactive memory management.
        namespace: The namespace structure for organizing memories in LangGraph's BaseStore.
            Uses runtime configuration with placeholders like `{langgraph_user_id}`.
        store: The BaseStore to use for searching. If not provided, the tool will use the configured BaseStore in your graph or entrypoint.
            Only set if you intend on using these tools outside the LangGraph context.

    Returns:
        memory_tool (Tool): A decorated async function that can be used as a tool for memory management.
            The tool supports creating, updating, and deleting memories with proper validation.

    The resulting tool has a signature that looks like the following:
        ```python
        from typing import Literal


        def manage_memory(
            content: str | None = None,  # Content for new/updated memory
            id: str | None = None,  # ID of existing memory to update/delete
            action: Literal["create", "update", "delete"] = "create",
        ) -> str: ...
        ```
        _Note: the tool supports both sync and async usage._

    !!! note "Namespace Configuration"
        The namespace is configured at runtime through the `config` parameter:
        ```python
        # Example: Per-user memory storage
        config = {"configurable": {"langgraph_user_id": "user-123"}}
        # Results in namespace: ("memories", "user-123")

        # Example: Team-wide memory storage
        config = {"configurable": {"langgraph_user_id": "team-x"}}
        # Results in namespace: ("memories", "team-x")
        ```

    Tip:
        This tool connects with the LangGraph [BaseStore](https://langchain-ai.github.io/langgraph/reference/store/#langgraph.store.base.BaseStore) configured in your graph or entrypoint.
        It will not work if you do not provide a store.

    !!! example "Examples"
        ```python
        from langmem import create_manage_memory_tool
        from langgraph.func import entrypoint
        from langgraph.store.memory import InMemoryStore

        memory_tool = create_manage_memory_tool(
            # All memories saved to this tool will live within this namespace
            # The brackets will be populated at runtime by the configurable values
            namespace=("project_memories", "{langgraph_user_id}"),
        )

        store = InMemoryStore(
            index={
                "dims": 1536,
                "embed": "openai:text-embedding-3-small",
            }
        )


        @entrypoint(store=store)
        async def workflow(state: dict, *, previous=None):
            # Other work....
            result = await memory_tool.ainvoke(state)
            print(result)
            return entrypoint.final(value=result, save={})


        config = {
            "configurable": {
                # This value will be formatted into the namespace you configured above ("project_memories", "{langgraph_user_id}")
                "langgraph_user_id": "123e4567-e89b-12d3-a456-426614174000"
            }
        }
        # Create a new memory
        await workflow.ainvoke(
            {"content": "Team prefers to use Python for backend development"},
            config=config,
        )
        # Output: 'created memory 123e4567-e89b-12d3-a456-426614174000'

        # Update an existing memory
        result = await workflow.ainvoke(
            {
                "id": "123e4567-e89b-12d3-a456-426614174000",
                "content": "Team uses Python for backend and TypeScript for frontend",
                "action": "update",
            },
            config=config,
        )
        print(result)
        # Output: 'updated memory 123e4567-e89b-12d3-a456-426614174000'
        ```

        You can use in LangGraph's prebuilt `create_react_agent`:

        ```python
        from langgraph.prebuilt import create_react_agent
        from langgraph.config import get_config, get_store

        def prompt(state):
            config = get_config()
            memories = get_store().search(
                # Search within the same namespace as the one
                # we've configured for the agent
                ("memories", config["configurable"]["langgraph_user_id"]),
            )
            system_prompt = f"\"\"You are a helpful assistant.
        <memories>
        {memories}
        </memories>
        \"\"\"
            system_message = {"role": "system", "content": system_prompt}
            return [system_message, *state["messages"]]

        agent = create_react_agent(
            "anthropic:claude-3-5-sonnet-latest",
            tools=[
                create_manage_memory_tool(namespace=("memories", "{langgraph_user_id}")),
            ],
            store=store,
        )

        agent.invoke(
            {"messages": [{"role": "user", "content": "We've decided we like golang more than python for backend work"}]},
            config=config,
        )
        ```


        If you want to customize the expected schema for memories, you can do so by providing a `schema` argument.
        ```python
        from pydantic import BaseModel


        class UserProfile(BaseModel):
            name: str
            age: int | None = None
            recent_memories: list[str] = []
            preferences: dict | None = None


        memory_tool = create_manage_memory_tool(
            # All memories saved to this tool will live within this namespace
            # The brackets will be populated at runtime by the configurable values
            namespace=("memories", "{langgraph_user_id}", "user_profile"),
            schema=UserProfile,
            actions_permitted=["create", "update"],
            instructions="Update the existing user profile (or create a new one if it doesn't exist) based on the shared information.",
        )
        store = InMemoryStore(
            index={
                "dims": 1536,
                "embed": "openai:text-embedding-3-small",
            }
        )
        agent = create_react_agent(
            "anthropic:claude-3-5-sonnet-latest",
            prompt=prompt,
            tools=[
                memory_tool,
            ],
            store=store,
        )

        result = agent.invoke(
            {
                "messages": [
                    {
                        "role": "user",
                        "content": "I'm 60 years old and have been programming for 5 days.",
                    }
                ]
            },
            config=config,
        )
        result["messages"][-1].pretty_print()
        # I've created a memory with your age of 60 and noted that you started programming 5 days ago...

        result = agent.invoke(
            {
                "messages": [
                    {"role": "user", "content": "Just had by 61'st birthday today!!"}
                ]
            },
            config=config,
        )
        result["messages"][-1].pretty_print()
        # Happy 61st birthday! 🎂 I've updated your profile to reflect your new age. Is there anything else I can help you with?
        print(
            store.search(
                ("memories", "123e4567-e89b-12d3-a456-426614174000", "user_profile")
            )
        )
        # [Item(
        #     namespace=['memories', '123e4567-e89b-12d3-a456-426614174000', 'user_profile'],
        #     key='1528553b-0900-4363-8dc2-c6b72844096e',
        #     value={
        # highlight-next-line
        #         'content': UserProfile(
        #             name='User',
        #             age=61,
        #             recent_memories=['Started programming 5 days ago'],
        #             preferences={'programming_experience': '5 days'}
        #         )
        #     },
        #     created_at='2025-02-07T01:12:14.383762+00:00',
        #     updated_at='2025-02-07T01:12:14.383763+00:00',
        #     score=None
        # )]
        ```

        If you want to limit the actions that can be taken by the tool, you can do so by providing a `actions_permitted` argument.

    """
    namespacer = utils.NamespaceTemplate(namespace)
    action_type = typing.Literal[*actions_permitted]
    assert actions_permitted, "actions_permitted cannot be empty"
    default_action = "create" if "create" in actions_permitted else actions_permitted[0]
    initial_store = store

    async def amanage_memory(
        content: typing.Optional[schema] = None,  # type: ignore
        action: action_type = default_action,  # type: ignore
        *,
        id: typing.Optional[uuid.UUID] = None,
    ):
        store = _get_store(initial_store)
        if action not in actions_permitted:
            raise ValueError(
                f"Invalid action {action}. Must be one of {actions_permitted}."
            )

        if action == "create" and id is not None:
            raise ValueError(
                "You cannot provide a MEMORY ID when creating a MEMORY. Please try again, omitting the id argument."
            )

        if action in ("delete", "update") and not id:
            raise ValueError(
                "You must provide a MEMORY ID when deleting or updating a MEMORY."
            )
        namespace = namespacer()
        if action == "delete":
            await store.adelete(namespace, key=str(id))
            return f"Deleted memory {id}"

        id = id or uuid.uuid4()
        await store.aput(
            namespace,
            key=str(id),
            value={"content": _ensure_json_serializable(content)},
        )
        return f"{action}d memory {id}"

    def manage_memory(
        content: typing.Optional[schema] = None,  # type: ignore
        action: action_type = "create",  # type: ignore
        *,
        id: typing.Optional[uuid.UUID] = None,
    ):
        store = _get_store(initial_store)
        if action == "create" and id is not None:
            raise ValueError(
                "You cannot provide a MEMORY ID when creating a MEMORY. Please try again, omitting the id argument."
            )

        if action in ("delete", "update") and not id:
            raise ValueError(
                "You must provide a MEMORY ID when deleting or updating a MEMORY."
            )
        namespace = namespacer()
        if action == "delete":
            store.delete(namespace, key=str(id))
            return f"Deleted memory {id}"

        id = id or uuid.uuid4()
        store.put(
            namespace,
            key=str(id),
            value={"content": _ensure_json_serializable(content)},
        )
        return f"{action}d memory {id}"

    description = """Create, update, or delete persistent MEMORIES to persist across conversations.
Include the MEMORY ID when updating or deleting a MEMORY. Omit when creating a new MEMORY - it will be created for you.
{instructions}""".format(instructions=instructions)

    return StructuredTool.from_function(
        manage_memory, amanage_memory, name=name, description=description
    )


_MEMORY_SEARCH_INSTRUCTIONS = ""


def create_search_memory_tool(
    namespace: tuple[str, ...] | str,
    *,
    instructions: str = _MEMORY_SEARCH_INSTRUCTIONS,
    store: BaseStore | None = None,
    response_format: typing.Literal["content", "content_and_artifact"] = "content",
    name: str = "search_memory",
):
    """Create a tool for searching memories stored in a LangGraph BaseStore.

    This function creates a tool that allows AI assistants to search through previously stored
    memories using semantic or exact matching. The tool returns both the memory contents and
    the raw memory objects for advanced usage.

    Args:
        instructions: Custom instructions for when to use the search tool.
            Defaults to a predefined set of guidelines.
        namespace: The namespace structure for organizing memories in LangGraph's BaseStore.
            Uses runtime configuration with placeholders like `{langgraph_user_id}`.
            See [Memory Namespaces](../concepts/conceptual_guide.md#memory-namespaces).
        store: The BaseStore to use for searching. If not provided, the tool will use the configured BaseStore in your graph or entrypoint.
            Only set if you intend on using these tools outside the LangGraph context.

    Returns:
        search_tool (Tool): A decorated function that can be used as a tool for memory search.
            The tool returns both serialized memories and raw memory objects.

    The resulting tool has a signature that looks like the following:
        ```python
        def search_memory(
            query: str,  # Search query to match against memories
            limit: int = 10,  # Maximum number of results to return
            offset: int = 0,  # Number of results to skip
            filter: dict | None = None,  # Additional filter criteria
        ) -> tuple[list[dict], list]: ...  # Returns (serialized memories, raw memories)
        ```
    _Note: the tool supports both sync and async usage._


    Tip:
        This tool connects with the LangGraph [BaseStore](https://langchain-ai.github.io/langgraph/reference/store/#langgraph.store.base.BaseStore) configured in your graph or entrypoint.
        It will not work if you do not provide a store.

    !!! example "Examples"
        ```python
        from langmem import create_search_memory_tool
        from langgraph.func import entrypoint
        from langgraph.store.memory import InMemoryStore

        search_tool = create_search_memory_tool(
            namespace=("project_memories", "{langgraph_user_id}"),
        )

        store = InMemoryStore(
            index={
                "dims": 1536,
                "embed": "openai:text-embedding-3-small",
            }
        )


        @entrypoint(store=store)
        async def workflow(state: dict, *, previous=None):
            # Search for memories about Python
            memories, _ = await search_tool.ainvoke(
                {"query": "Python preferences", "limit": 5}
            )
            print(memories)
            return entrypoint.final(value=memories, save={})
        ```
    """
    namespacer = utils.NamespaceTemplate(namespace)
    initial_store = store

    async def asearch_memory(
        query: str,
        *,
        limit: int = 10,
        offset: int = 0,
        filter: typing.Optional[dict] = None,
    ):
        store = _get_store(initial_store)
        namespace = namespacer()
        memories = await store.asearch(
            namespace,
            query=query,
            filter=filter,
            limit=limit,
            offset=offset,
        )
        if response_format == "content_and_artifact":
            return [m.dict() for m in memories], memories
        return [m.dict() for m in memories]

    def search_memory(
        query: str,
        *,
        limit: int = 10,
        offset: int = 0,
        filter: typing.Optional[dict] = None,
    ):
        store = _get_store(initial_store)
        namespace = namespacer()
        memories = store.search(
            namespace,
            query=query,
            filter=filter,
            limit=limit,
            offset=offset,
        )
        if response_format == "content_and_artifact":
            return [m.dict() for m in memories], memories
        return [m.dict() for m in memories]

    description = """Search your long-term memories for information relevant to your current context. {instructions}""".format(
        instructions=instructions
    )

    return StructuredTool.from_function(
        search_memory,
        asearch_memory,
        name=name,
        description=description,
        response_format=response_format,
    )


def _get_store(initial_store: BaseStore | None = None) -> BaseStore:
    try:
        if initial_store is not None:
            store = initial_store
        else:
            store = get_store()
        return store
    except RuntimeError as e:
        raise errors.ConfigurationError("Could not get store") from e


def _ensure_json_serializable(content: typing.Any) -> typing.Any:
    # Right now just support primitives and pydantic models
    if isinstance(content, (str, int, float, bool, dict, list)):
        return content
    if hasattr(content, "model_dump"):
        try:
            return content.model_dump(mode="json")
        except Exception as e:
            logger.error(e)
            return str(content)
    return content
